from typing import Dict

import pandas as pd
from openghg.types import SurfaceTypes

# from openghg.types import DataTypes  # Would this be more appropriate?
# This does include Footprint as well as the input obs data types?


def _extract_site_param(data_type: str) -> Dict:
    """
    Extract site data for a given data type based on the available data
    file.

    For data types:
    - "GCWERKS", "CRDS"
       - "process_gcwerks_parameters.json"
    - "NPL", "BTT", "THAMESBARRIER"
       - "lghg_sites.json"
    - "BEACO2N"
       - "beaco2n_site_data.json"
    - Other
       - none specified

    Args:
        data_type (str) : Accepted data type (defined in SurfaceTypes)

    Returns:
        Dict: Details of site params for the data type extracted from json file

        If no file is available for data_type input:
            empty dictionary

    TODO: Create files of site details for other data types and incorporate here.
    """
    from openghg.util import load_json

    data_type = data_type.upper()

    lghg_sites = {"NPL": "NPL", "BTT": "BTT", "THAMESBARRIER": "TMB"}

    # TODO: Could create another json / csv which includes links
    # between GCWERKS and instruments (with a warning) if needed

    if data_type == "GCWERKS":
        # "process_gcwerks_parameters.json" - ["GCWERKS"]["sites"], sites are keys
        params_full = load_json(filename="process_gcwerks_parameters.json")
        site_params: Dict = params_full[data_type]["sites"]
    elif data_type == "CRDS":
        # "process_gcwerks_parameters.json" - ["CRDS"]["sites"], sites are keys
        params_full = load_json(filename="process_gcwerks_parameters.json")
        crds_params = params_full[data_type]["sites"]
        site_params = {key: value for key, value in crds_params.items() if len(key) == 3 and key.isupper()}
    elif data_type in lghg_sites.keys():
        # TODO: May want to update this now new sites have been added to lghg_data.json
        # Do these sites have their own data types?
        # "lghg_data.json" - ["sites"], sites are keys
        params_full = load_json(filename="lghg_data.json")
        site_name = lghg_sites[data_type]
        site_params = {}
        site_params[site_name] = params_full["sites"][site_name]
    elif data_type == "BEACO2N":
        # "beaco2n_site_data.json" - sites are keys
        params_full = load_json(filename="beaco2n_site_data.json")
        site_params = params_full
    else:
        site_params = {}

    return site_params


def _extract_site_names(site_params: Dict, data_type: str) -> Dict:
    """
    Extracts long names from site parameters - expects output to
    match format from `_extract_site_param()` function.

    Args:
        site_params (dict) : Dictionary of site data (extracted from
        relevant json file)
        data_type (str) : Associated data type for this data

    Returns:
        Dict: Long names for each site code
    """
    if data_type in ("GCWERKS", "CRDS"):
        name = "gcwerks_site_name"
    else:
        name = "long_name"

    site_names = {}
    for site, data in site_params.items():
        site_names[site] = data[name]

    return site_names


def summary_data_types() -> pd.DataFrame:
    """
    Create summary DataFrame of accepted input data types. This includes
    the site code, long name, platform and data type.

    Returns:
        pandas.DataFrame

    TODO: Add data_type details for mobile / column etc. when added
    """
    # Could include input for surface / mobile / column etc.?
    surface_data_types = list(SurfaceTypes.__members__)

    collated_site_data = pd.DataFrame()

    for data_type in surface_data_types:
        site_params = _extract_site_param(data_type)

        if site_params:
            site_codes = list(site_params.keys())
            site_names_dict = _extract_site_names(site_params, data_type)
            site_names = [site_names_dict[code] for code in site_codes]
        else:
            site_codes = [""]
            site_names = [""]

        site_data = pd.DataFrame({"Site code": site_codes, "Long name": site_names})
        site_data["Data type"] = data_type
        site_data["Platform"] = "surface site"

        collated_site_data = pd.concat([collated_site_data, site_data], ignore_index=True)

    # TODO: May want to sort by the site code but removed for now as nice
    # to keep GCWERKS and CRDS first.
    # collated_site_data = collated_site_data.sort_values(by="Site code")

    return collated_site_data



from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

import numpy as np
from xarray import Dataset

__all__ = ["DataSchema"]


@dataclass
class DataSchema:
    """
    Create schema for a data type based on inputs for components of the data.
    Expected format is based against an xarray Dataset.

    Example inputs:
        DataSchema(
            data_vars = {"fp": ("time", "lat", "lon"), ...},
            dtypes = {"fp" : np.floating, ...},
            dims = ["time", "lat", "lon", ...]
        )
    """

    # TODO : Change or add additional checks as needed

    data_vars: Optional[Dict[str, Tuple[str, ...]]] = None
    dtypes: Optional[Dict[str, type]] = None
    dims: Optional[List[str]] = None

    def _check_data_vars(self, data: Dataset) -> None:
        """
        Check data variables and their dimensions of data against the schema.

        Args:
            data : xarray Dataset to be checked
        Returns:
            None

            Raises a ValueError with details if expected data variable is
            not present or expected dimensions are not present for a data variable
            based on the DataSchema object.
        """
        expected_data_vars = self.data_vars
        if expected_data_vars is None:
            print("No data variables to check against schema")
            return None
        else:
            expected_dv = expected_data_vars.keys()

        data_vars = data.data_vars

        for edv in expected_dv:
            if edv in data_vars:
                dims = data[edv].dims
                expected_dv_dims = expected_data_vars[edv]
                for edim in expected_dv_dims:
                    if edim not in dims:
                        raise ValueError(
                            f"Missing dimension for data variable: {edv}, {edim}. Current dims: {dims}"
                        )
            else:
                raise ValueError(f"Expected data variable: {edv} not present in standardised data")

    def _check_dims(self, data: Dataset) -> None:
        """
        Check dimensions of data against the the schema.

        Args:
            data : xarray Dataset to be checked
        Returns:
            None

            Raises a ValueError with details if expected dimensions
            are not present in data based on the DataSchema object.
        """
        expected_dims = self.dims
        if expected_dims is None:
            print("No dims to check against schema")
            return None

        dims = data.dims

        for edim in expected_dims:
            if edim not in dims:
                raise ValueError(f"Expected dimension: {edim} not present in standardised data")

    def _check_dtypes(self, data: Dataset) -> None:
        """
        Check dtypes of variables and coordinates of data against the schema.

        Args:
            data : xarray Dataset to be checked
        Returns:
            None

            Raises a ValueError with details if data variables and coordinates
            are not of expected data types based on the DataSchema object.
        """
        expected_data_types = self.dtypes
        if expected_data_types is None:
            print("No data types to check against schema")
            return None

        for variable, edata_type in expected_data_types.items():
            if variable in data:
                dtype = data[variable].dtype
                if not np.issubdtype(dtype, edata_type):
                    raise ValueError(
                        f"Expected data type of variable {variable} to be: {edata_type}. Current {dtype}"
                    )

    def validate_data(
        self,
        data: Dataset,
    ) -> None:
        """
        Validate input data based on schema.

        Currently check can include:
         - data variables are present with expected dimensions.
         - general dimensions are present
         - data types of data variables and coordinates match to expected values

        Args:
            data : xarray Dataset to be validated
        Returns:
            None

            Raises a ValueError with details if the input data does not adhere
            to the defined DataSchema.
        """

        if self.data_vars is not None:
            self._check_data_vars(data)

        if self.dims is not None:
            self._check_dims(data)

        if self.dtypes is not None:
            self._check_dtypes(data)

    # TODO: Add string method for pretty printing



"""
The ModelScenario class allows users to collate related data sources and calculate
modelled output based on this data. The types of data currently included are:
 - Timeseries observation data (ObsData)
 - Fixed domain sensitivity maps known as footprints (FootprintData)
 - Fixed domain flux maps (FluxData) - multiple maps can be included and
 referenced by source name

TODO: Also need to incorporate boundary conditions

A ModelScenario instance can be created by searching the object store manually
and providing these outputs:
>>> obs = get_obs_surface(site, species, ...)
>>> footprint = get_footprint(site, domain, ...)
>>> flux = get_flux(species, source, ...)
>>> model = ModelScenario(obs=obs, footprint=footprint, flux=flux)

A ModelScenario instance can also be created using keywords to search the object store:
>>> model = ModelScenario(site,
                          species,
                          inlet,
                          network,
                          domain,
                          sources=sources,
                          start_date=start_date,
                          end_date=end_date)

A ModelScenario instance can also be initialised and then populated after creation:
>>> model = ModelScenario()
>>> model.add_obs(obs=obs)
>>> model.add_footprint(site, inlet, domain, ...)
>>> model.add_flux(species, domain, sources, ...)

Once created, methods can be called on ModelScenario which will combine these
data sources and cache the outputs (if requested) to make for quicker calculation.

>>> modelled_obs = model.calc_modelled_obs()
>>> combined_data = model.footprints_data_merge()

If some input types needed for these operations are missing, the user will be alerted
on which data types are missing.
"""

from typing import Any, Dict, List, Literal, Optional, Sequence, Tuple, Union, cast

import numpy as np
from openghg.dataobjects import BoundaryConditionsData, FluxData, FootprintData, ObsData
from openghg.retrieve import get_bc, get_flux, get_footprint, get_obs_surface, search
from openghg.util import synonyms
from pandas import Timestamp
from xarray import DataArray, Dataset

__all__ = ["ModelScenario", "combine_datasets", "stack_datasets", "calc_dim_resolution", "match_dataset_dims"]


# TODO: Really with the emissions, they shouldn't need to match against a domain
# We should be able to grab global/bigger area emissions and cut that down
# to whichever area our LPDM model covers.

# TODO: Add static methods for different ways of creating the class
# e.g. from_existing_data(), from_search(), empty() , ...

ParamType = Union[List[Dict[str, Optional[str]]], Dict[str, Optional[str]]]
methodType = Optional[Literal["nearest", "pad", "ffill", "backfill", "bfill"]]


class ModelScenario:
    """
    This class stores together observation data with ancillary data and allows
    operations to be performed combining these inputs.
    """

    def __init__(
        self,
        site: Optional[str] = None,
        species: Optional[str] = None,
        inlet: Optional[str] = None,
        network: Optional[str] = None,
        domain: Optional[str] = None,
        model: Optional[str] = None,
        metmodel: Optional[str] = None,
        source: Optional[str] = None,
        sources: Optional[Union[str, Sequence]] = None,
        bc_input: Optional[str] = None,
        start_date: Optional[Union[str, Timestamp]] = None,
        end_date: Optional[Union[str, Timestamp]] = None,
        obs: Optional[ObsData] = None,
        footprint: Optional[FootprintData] = None,
        flux: Optional[Union[FluxData, Dict[str, FluxData]]] = None,
        bc: Optional[BoundaryConditionsData] = None,
    ):
        """
        Create a ModelScenario instance based on a set of keywords to be
        or directly supplied objects. This can be created as an empty class to be
        populated.

        The keywords are related to observation, footprint and flux data
        which may be available within the object store. The combination of these supplied
        will be used to extract the relevant data. Related keywords are as follows:
         - Observation data: site, species, inlet, network, start_date, end_data
         - Footprint data: site, inlet, domain, model, metmodel, species, start_date, end_date
         - Flux data: species, sources, domain, start_date, end_date

        Args:
            site : Site code e.g. "TAC"
            species : Species code e.g. "ch4"
            inlet : Inlet value e.g. "10m"
            network : Network name e.g. "AGAGE"
            domain : Domain name e.g. "EUROPE"
            model : Model name used in creation of footprint e.g. "NAME"
            metmodel : Name of met model used in creation of footprint e.g. "UKV"
            sources : Emissions sources
            bc_input : Input keyword for boundary conditions e.g. "mozart" or "cams"
            start_date : Start of date range to use. Note for flux this may not be applied
            end_date : End of date range to use. Note for flux this may not be applied
            obs : Supply ObsData object directly (e.g. from get_obs...() functions)
            footprint : Supply FootprintData object directly (e.g. from get_footprint() function)
            flux : Supply FluxData object directly (e.g. from get_flux() function)

        Returns:
            None

            Sets up instance of class with associated values.

        TODO: For obs, footprint, flux should we also allow Dataset input and turn
        these into the appropriate class?
        """

        self.obs: Optional[ObsData] = None
        self.footprint: Optional[FootprintData] = None
        self.fluxes: Optional[Dict[str, FluxData]] = None
        self.bc: Optional[BoundaryConditionsData] = None

        if species is not None:
            species = synonyms(species)

        # Add observation data (directly or through keywords)
        self.add_obs(
            site=site,
            species=species,
            inlet=inlet,
            network=network,
            start_date=start_date,
            end_date=end_date,
            obs=obs,
        )

        # Make sure obs data is present, make sure inputs match to metadata
        if self.obs is not None:
            obs_metadata = self.obs.metadata
            site = obs_metadata["site"]
            species = obs_metadata["species"]
            inlet = obs_metadata["inlet"]
            print("Updating any inputs based on observation data")
            print(f"site: {site}, species: {species}, inlet: {inlet}")

        # Add footprint data (directly or through keywords)
        self.add_footprint(
            site=site,
            inlet=inlet,
            domain=domain,
            model=model,
            metmodel=metmodel,
            start_date=start_date,
            end_date=end_date,
            species=species,
            footprint=footprint,
        )

        # Add flux data (directly or through keywords)
        self.add_flux(
            species=species,
            domain=domain,
            source=source,
            sources=sources,
            start_date=start_date,
            end_date=end_date,
            flux=flux,
        )

        # Add boundary conditions (directly or through keywords)
        self.add_bc(
            species=species,
            bc_input=bc_input,
            domain=domain,
            start_date=start_date,
            end_date=end_date,
            bc=bc,
        )

        # Initialise attributes used for caching
        self.scenario: Optional[Dataset] = None
        self.modelled_obs: Optional[DataArray] = None
        self.modelled_baseline: Optional[DataArray] = None
        self.flux_stacked: Optional[Dataset] = None

        # TODO: Check species, site etc. values align between inputs?

    def _get_data(self, keywords: ParamType, input_type: str) -> Any:
        """
        Use appropriate get function to search for data in object store.
        """

        get_functions = {
            "obs_surface": get_obs_surface,
            "footprint": get_footprint,
            "flux": get_flux,
            "boundary_conditions": get_bc,
        }

        # TODO: Add/write footprint and flux search? What's the syntax?
        search_functions = {"obs_surface": search}

        get_fn = get_functions[input_type]
        search_fn = search_functions.get(input_type)

        if isinstance(keywords, dict):
            keywords = [keywords]

        num_checks = len(keywords)
        for i, keyword_set in enumerate(keywords):
            try:
                data = get_fn(**keyword_set)  # type:ignore
            except (ValueError, AttributeError):
                num = i + 1
                if num == num_checks:
                    print(f"Unable to add {input_type} data based on keywords supplied.")
                    print(" Inputs - \n")
                    for key, value in keyword_set.items():
                        print(f" {key}: {value}\n")
                    if search_fn is not None:
                        data_search = search_fn(**keyword_set)
                        print("---- Search results ---")
                        print(data_search)
                        # TODO: If we can determine how many results are returned from search
                        # we can use this to give better information about why no data has
                        # been found for these inputs.
                data = None
            else:
                print(f"Adding {input_type} to model scenario")
                break

        return data

    def add_obs(
        self,
        site: Optional[str] = None,
        species: Optional[str] = None,
        inlet: Optional[str] = None,
        network: Optional[str] = None,
        start_date: Optional[Union[str, Timestamp]] = None,
        end_date: Optional[Union[str, Timestamp]] = None,
        obs: Optional[ObsData] = None,
    ) -> None:
        """
        Add observation data based on keywords or direct ObsData object.
        """
        from openghg.util import clean_string

        # Search for obs data based on keywords
        if site is not None and obs is None:
            site = clean_string(site)
            # search for obs based on suitable keywords - site, species, inlet
            obs_keywords = {
                "site": site,
                "species": species,
                "inlet": inlet,
                "network": network,
                "start_date": start_date,
                "end_date": end_date,
            }

            obs = self._get_data(obs_keywords, input_type="obs_surface")

        self.obs = obs

        # Add keywords to class for convenience
        if self.obs is not None:
            self.site = self.obs.metadata["site"]
            self.species = self.obs.metadata["species"]

    def add_footprint(
        self,
        site: Optional[str] = None,
        inlet: Optional[str] = None,
        domain: Optional[str] = None,
        model: Optional[str] = None,
        metmodel: Optional[str] = None,
        start_date: Optional[Union[str, Timestamp]] = None,
        end_date: Optional[Union[str, Timestamp]] = None,
        species: Optional[str] = None,
        footprint: Optional[FootprintData] = None,
    ) -> None:
        """
        Add footprint data based on keywords or direct FootprintData object.
        """
        from openghg.util import clean_string, species_lifetime

        # Search for footprint data based on keywords
        # - site, domain, inlet (can extract from obs), model, metmodel
        if site is not None and footprint is None:
            site = clean_string(site)
            if inlet is None and self.obs is not None:
                inlet = self.obs.metadata["inlet"]

            # TODO: Add case to deal with "multiple" inlets
            if inlet == "multiple":
                raise ValueError(
                    "Unable to deal with multiple inlets yet:\n Please change date range or specify a specific inlet"
                )

            footprint_keywords = {
                "site": site,
                "height": inlet,
                "domain": domain,
                "model": model,  # Not currently used in get_footprint - should be added
                # "metmodel": metmodel,  # Should be added to inputs for get_footprint()
                "start_date": start_date,
                "end_date": end_date,
            }

            # Check whether general inert footprint should be extracted (suitable for long-lived species)
            # or species specific footprint
            #  - needed for short-lived species (includes additional parameters for age of particles)
            #  - needed for carbon dioxide (include high time resolution footprint)
            species_lifetime_value = species_lifetime(species)
            if species_lifetime_value is not None or species == "co2":
                footprint_keywords["species"] = species

            footprint = self._get_data(footprint_keywords, input_type="footprint")

        self.footprint = footprint

        if self.footprint is not None and not hasattr(self, "site"):
            self.site = self.footprint.metadata["site"]

    def add_flux(
        self,
        species: Optional[str] = None,
        domain: Optional[str] = None,
        source: Optional[str] = None,
        sources: Optional[Union[str, Sequence]] = None,
        start_date: Optional[Union[str, Timestamp]] = None,
        end_date: Optional[Union[str, Timestamp]] = None,
        flux: Optional[Union[FluxData, Dict[str, FluxData]]] = None,
    ) -> None:
        """
        Add flux data based on keywords or direct FluxData object.
        Can add flux datasets for multiple sources.
        """
        if self.fluxes is not None:
            # Check current species in any flux data
            if species is not None:
                current_flux_1 = list(self.fluxes.values())[0]
                current_species = current_flux_1.metadata["species"]
                if species != current_species:
                    raise ValueError(
                        f"New data must match current species {current_species} in ModelScenario. Input value: {species}"
                    )

        if species is not None and flux is None:
            flux = {}

            if sources is None and source is not None:
                sources = [source]
            elif sources is None or isinstance(sources, str):
                sources = [sources]

            for name in sources:
                flux_keywords_1 = {"species": species, "source": name, "domain": domain}

                # For CO2 we need additional emissions data before a start_date to
                # match to high time resolution footprints.
                # For now, just extract all data
                if species == "co2":
                    flux_keywords = [flux_keywords_1]
                else:
                    flux_keywords_2 = flux_keywords_1.copy()

                    flux_keywords_1["start_date"] = start_date
                    flux_keywords_1["end_date"] = end_date

                    flux_keywords = [flux_keywords_1, flux_keywords_2]

                # TODO: Add something to allow for e.g. global domain or no domain

                flux_source = self._get_data(flux_keywords, input_type="flux")
                # TODO: May need to update this check if flux_source is empty FluxData() object
                if flux_source is not None:
                    flux[name] = flux_source

        elif flux is not None:
            if not isinstance(flux, dict):
                name = flux.metadata["source"]
                flux = {name: flux}

        # TODO: Make this so flux.anthro can be called etc. - link in some way
        if self.fluxes is not None:
            if flux is not None:
                self.fluxes.update(flux)
        else:
            self.fluxes = flux

        if self.fluxes is not None:
            if not hasattr(self, "species"):
                flux_values = list(self.fluxes.values())
                flux_1 = flux_values[0]
                self.species = flux_1.metadata["species"]

            self.flux_sources = list(self.fluxes.keys())

    def add_bc(
        self,
        species: Optional[str] = None,
        bc_input: Optional[str] = None,
        domain: Optional[str] = None,
        start_date: Optional[Union[str, Timestamp]] = None,
        end_date: Optional[Union[str, Timestamp]] = None,
        bc: Optional[BoundaryConditionsData] = None,
    ) -> None:
        """
        Add boundary conditions data based on keywords or direct BoundaryConditionsData object.
        """

        # Search for boundary conditions data based on keywords
        # - domain, species, bc_input
        if domain is not None and bc is None:

            bc_keywords = {
                "species": species,
                "domain": domain,
                "bc_input": bc_input,
                "start_date": start_date,
                "end_date": end_date,
                "species": species,
            }

            bc = self._get_data(bc_keywords, input_type="boundary_conditions")

        self.bc = bc

    def _check_data_is_present(self, need: Optional[Union[str, Sequence]] = None) -> None:
        """
        Check whether correct data types have been included. This should
        be used by functions to check whether they can perform the requested
        operation with the data types available.

        Args:
            need (list) : Names of objects needed for the function being called.
            Should be one or more of "obs", "footprint", "fluxes" (or "flux")

        Returns:
            None

            Raises ValueError is necessary data is missing.
        """
        if need is None:
            need = ["obs", "footprint"]
        elif isinstance(need, str):
            need = [need]

        need = ["fluxes" if value == "flux" else value for value in need]  # Make sure attributes match
        missing = []
        for attr in need:
            value = getattr(self, attr)
            if value is None:
                missing.append(attr)

                print(f"Must have {attr} data linked to this ModelScenario to run this function")
                print("Include this by using the add function, with appropriate inputs:")
                print("  ModelScenario.add_{attr}(...)")

        if missing:
            raise ValueError(f"Missing necessary {' and '.join(missing)} data.")

    def _get_platform(self) -> Optional[str]:
        """
        Find the platform for a site, if present.

        This will access the "acrg_site_info.json" file to find this information.
        """
        from openghg.util import load_json

        try:
            site = self.site
            site_upper = site.upper()
        except AttributeError:
            return None
        else:
            site_info = load_json(filename="acrg_site_info.json")
            try:
                site_details = site_info[site_upper]
            except KeyError:
                return None
            else:
                platform: str = site_details.get("platform")
                return platform

    def _align_obs_footprint(self, resample_to: str = "coarsest", platform: Optional[str] = None) -> Tuple:
        """
        Slice and resample obs and footprint data to align along time

        This slices the date to the smallest time frame
        spanned by both the footprint and obs, using the sliced start date
        The time dimension is resampled based on the resample_to input using the mean.
        The resample_to options are:
         - "coarsest" - resample to the coarsest resolution between obs and footprints
         - "obs" - resample to observation data frequency
         - "footprint" - resample to footprint data frequency
         - a valid resample period e.g. "2H"

        Args:
            resample_to: Resample option to use: either data based or using a valid pandas resample period.
            platform: Observation platform used to decide whether to resample

        Returns:
            tuple: Two xarray.Dataset with aligned time dimensions
        """
        from pandas import Timedelta

        # Check data is present (not None) and cast to correct type
        self._check_data_is_present(need=["obs", "footprint"])
        obs = cast(ObsData, self.obs)
        footprint = cast(FootprintData, self.footprint)

        obs_data = obs.data
        footprint_data = footprint.data

        resample_keyword_choices = ("obs", "footprint", "coarsest")

        # Check whether resample has been requested by specifying a specific period rather than a keyword
        if resample_to in resample_keyword_choices:
            force_resample = False
        else:
            force_resample = True

        if platform is not None:
            platform = platform.lower()
            # Do not apply resampling for "satellite" (but have re-included "flask" for now)
            if platform == "satellite":
                return obs_data, footprint_data

        # Whether sampling period is present or we need to try to infer this
        infer_sampling_period = False
        # Get the period of measurements in time
        obs_attributes = obs_data.attrs
        if "averaged_period" in obs_attributes:
            obs_data_period_s = float(obs_attributes["averaged_period"])
        elif "sampling_period" in obs_attributes:
            sampling_period = obs_attributes["sampling_period"]
            if sampling_period == "NOT_SET":
                infer_sampling_period = True
            else:
                obs_data_period_s = float(sampling_period)
            obs_data_period_s = float(obs_attributes["sampling_period"])
        elif "sampling_period_estimate" in obs_attributes:
            estimate = obs_attributes["sampling_period_estimate"]
            print(f"WARNING: Using estimated sampling period of {estimate}s for observational data")
            obs_data_period_s = float(estimate)
        else:
            infer_sampling_period = True

        if infer_sampling_period:
            # Attempt to derive sampling period from frequency of data
            obs_data_period_s = np.nanmedian(
                (obs_data.time.data[1:] - obs_data.time.data[0:-1]) / 1e9
            ).astype("float32")

            obs_data_period_s_min = np.diff(obs_data.time.data).min() / 1e9
            obs_data_period_s_max = np.diff(obs_data.time.data).max() / 1e9

            max_diff = (obs_data_period_s_max - obs_data_period_s_min).astype(float)

            # Check if the periods differ by more than 1 second
            if max_diff > 1.0:
                raise ValueError("Sample period can be not be derived from observations")

        # TODO: Check regularity of the data - will need this to decide is resampling
        # is appropriate or need to do checks on a per time point basis

        obs_data_period_ns = obs_data_period_s * 1e9
        obs_data_timeperiod = Timedelta(obs_data_period_ns, unit="ns")

        # Derive the footprints period from the frequency of the data
        footprint_data_period_ns = np.nanmedian(
            (footprint_data.time.data[1:] - footprint_data.time.data[0:-1]).astype("int64")
        )
        footprint_data_timeperiod = Timedelta(footprint_data_period_ns, unit="ns")

        # If resample_to is set to "coarsest", check whether "obs" or "footprint" have lower resolution
        if resample_to == "coarsest":
            if obs_data_timeperiod >= footprint_data_timeperiod:
                resample_to = "obs"
            elif obs_data_timeperiod < footprint_data_timeperiod:
                resample_to = "footprint"

        # Here we want timezone naive Timestamps
        # Add sampling period to end date to make sure resample includes these values when matching
        obs_startdate = Timestamp(obs_data.time[0].values)
        obs_enddate = Timestamp(obs_data.time[-1].values) + obs_data_timeperiod
        footprint_startdate = Timestamp(footprint_data.time[0].values)
        footprint_enddate = Timestamp(footprint_data.time[-1].values) + footprint_data_timeperiod

        start_date = max(obs_startdate, footprint_startdate)
        end_date = min(obs_enddate, footprint_enddate)

        # Ensure lower range is covered for obs
        start_obs_slice = start_date - Timedelta("1ns")
        # Ensure extra buffer is added for footprint based on fp timeperiod.
        # This is to ensure footprint can be forward-filled to obs (in later steps)
        start_footprint_slice = start_date - (footprint_data_timeperiod - Timedelta("1ns"))
        # Subtract very small time increment (1 nanosecond) to make this an exclusive selection
        end_slice = end_date - Timedelta("1ns")

        obs_data = obs_data.sel(time=slice(start_obs_slice, end_slice))
        footprint_data = footprint_data.sel(time=slice(start_footprint_slice, end_slice))

        # Only non satellite datasets with different periods need to be resampled
        timeperiod_diff_s = np.abs(obs_data_timeperiod - footprint_data_timeperiod).total_seconds()
        tolerance = 1e-9  # seconds

        if timeperiod_diff_s >= tolerance or force_resample:
            base = start_date.hour + start_date.minute / 60.0 + start_date.second / 3600.0

            if resample_to == "obs":
                resample_period = str(round(obs_data_timeperiod / np.timedelta64(1, "h"), 5)) + "H"
                footprint_data = footprint_data.resample(indexer={"time": resample_period}, base=base).mean()

            elif resample_to == "footprint":
                resample_period = str(round(footprint_data_timeperiod / np.timedelta64(1, "h"), 5)) + "H"
                obs_data = obs_data.resample(indexer={"time": resample_period}, base=base).mean()

            else:
                resample_period = resample_to
                footprint_data = footprint_data.resample(indexer={"time": resample_period}, base=base).mean()
                obs_data = obs_data.resample(indexer={"time": resample_period}, base=base).mean()

        return obs_data, footprint_data

    def combine_obs_footprint(
        self,
        resample_to: str = "coarsest",
        platform: Optional[str] = None,
        cache: bool = True,
        recalculate: bool = False,
    ) -> Dataset:
        """
        Combine observation and footprint data so these are on the same time
        axis. This will both slice and resample the data to align this axis.

        - Data is slices to smallest timeframe spanned by both footprint and obs
        - Data is resampled according to resample_to input and using the mean
        - Data is combined into one dataset

        Args:
            resample_to: Resample option to use for averaging:
                          - either one of ["coarsest", "obs", "footprint"] to match to the datasets
                          - or using a valid pandas resample period e.g. "2H".
                         Default = "coarsest".
            platform: Observation platform used to decide whether to resample
            cache: Cache this data after calculation. Default = True.

        Returns:
            xarray.Dataset: Combined dataset aligned along the time dimension

            If cache is True:
                This data will be also be cached as the ModelScenario.scenario attribute.
        """

        self._check_data_is_present(need=["obs", "footprint"])
        obs = cast(ObsData, self.obs)
        footprint = cast(FootprintData, self.footprint)

        # Return any matching cached data
        if self.scenario is not None and not recalculate:
            if self.scenario.attrs["resample_to"] == resample_to:
                return self.scenario

        # As we're not processing any satellite data yet just set tolerance to None
        tolerance = None
        if platform is None:
            platform = self._get_platform()

        # Align and merge the observation and footprint Datasets
        aligned_obs, aligned_footprint = self._align_obs_footprint(resample_to=resample_to, platform=platform)
        combined_dataset = combine_datasets(
            dataset_A=aligned_obs, dataset_B=aligned_footprint, tolerance=tolerance
        )

        # Transpose to keep time in the last dimension position in case it has been moved in resample
        combined_dataset = combined_dataset.transpose(..., "time")

        # Save the observation data units
        try:
            mf = obs.data["mf"]
            units: Optional[float] = float(mf.attrs["units"])
        except KeyError:
            units = None
        except AttributeError:
            raise AttributeError("Unable to read mf attribute from observation data.")

        if units is not None:
            combined_dataset.update({"fp": (combined_dataset.fp.dims, (combined_dataset["fp"].data / units))})
            if self.species == "co2":
                combined_dataset.update(
                    {
                        "fp_HiTRes": (
                            combined_dataset.fp_HiTRes.dims,
                            (combined_dataset.fp_HiTRes.data / units),
                        )
                    }
                )

        attributes = {}
        attributes_obs = obs.data.attrs
        attributes_footprint = footprint.data.attrs
        attributes.update(attributes_footprint)
        attributes.update(attributes_obs)

        attributes["resample_to"] = resample_to

        combined_dataset.attrs.update(attributes)

        if cache:
            self.scenario = combined_dataset

        return combined_dataset

    def _clean_sources_input(self, sources: Optional[Union[str, List]] = None) -> List:
        """
        Check sources input and make sure this is a list. If None, this will extract
        all sources from self.fluxes.
        """
        self._check_data_is_present(need=["fluxes"])
        flux_dict = cast(Dict[str, FluxData], self.fluxes)

        if sources is None:
            sources = list(flux_dict.keys())
        elif isinstance(sources, str):
            sources = [sources]

        return sources

    def combine_flux_sources(
        self, sources: Optional[Union[str, List]] = None, cache: bool = True, recalculate: bool = False
    ) -> Dataset:
        """
        Combine together flux sources on the time dimension. This will align to
        the time of the highest frequency flux source both for time range and frequency.

        Args:
            sources : Names of sources to combine. Should already be attached to ModelScenario.
            cache : Cache this data after calculation. Default = True

        Returns:
            Dataset: All flux sources stacked on the time dimension.
        """
        self._check_data_is_present(need=["fluxes"])
        flux_dict = cast(Dict[str, FluxData], self.fluxes)

        time_dim = "time"

        sources = self._clean_sources_input(sources)
        sources_str = ", ".join(sources)

        # Return any matching cached data
        if self.flux_stacked is not None and not recalculate:
            if self.flux_stacked.attrs["sources"] == sources_str:
                return self.flux_stacked

        flux_datasets = [flux_dict[source].data for source in sources]

        if len(sources) == 1:
            return flux_datasets[0]

        # Make sure other dimensions than time are aligned between flux datasets
        # - expects values to be closely aligned so only allows for a small floating point tolerance
        dims = list(flux_datasets[0].dims)
        dims.remove("time")
        flux_datasets = match_dataset_dims(flux_datasets, dims=dims)

        try:
            flux_stacked = stack_datasets(flux_datasets, dim=time_dim, method="ffill")
        except ValueError:
            raise ValueError(f"Unable to combine flux data for sources: {sources_str}")

        if cache:
            flux_stacked.attrs["sources"] = sources_str
            self.flux_stacked = flux_stacked

        return flux_stacked

    def _check_footprint_resample(self, resample_to: str) -> Dataset:
        """
        Check whether footprint needs resampling based on resample_to input.
        Ignores resample_to keywords of ("coarsest", "obs", "footprint") as this is
        for comparison with observation data but uses pandas frequencies to resample.
        """
        footprint = cast(FootprintData, self.footprint)

        if resample_to in ("coarsest", "obs", "footprint"):
            return footprint.data
        else:
            footprint_data = footprint.data
            time = footprint_data["time"].values
            start_date = Timestamp(time[0])
            base = start_date.hour + start_date.minute / 60.0 + start_date.second / 3600.0
            footprint_data = footprint_data.resample(indexer={"time": resample_to}, base=base).mean()
            return footprint_data

    def _param_setup(
        self,
        param: str = "modelled_obs",
        resample_to: str = "coarsest",
        platform: Optional[str] = None,
        recalculate: bool = False,
    ) -> bool:
        """
        Decide if calculation is needed for input parameter and set up
        underlying parameters accordingly. This will populate the
        self.scenario attribute if not already present or if this needs
        to be recalculated.

        Args:
            param : Name of the parameter being calculated.
                    Should be one of "modelled_obs", "modelled_baseline"
            resample_to: Resample option to use for averaging:
                          - either one of ["coarsest", "obs", "footprint"] to match to the datasets
                          - or using a valid pandas resample period e.g. "2H".
                         Default = "coarsest".
            platform: Observation platform used to decide whether to resample e.g. "site", "satellite".
            cache: Cache this data after calculation. Default = True.
            recalculate: Make sure to recalculate this data rather than return from cache. Default = False.

        Returns:
            bool: True if param should be calculated, False otherwise

            Populates details of ModelScenario.scenario to use in calculation.
        """

        try:
            parameter = getattr(self, param)
        except AttributeError:
            raise ValueError(f"Did not recognise input for {param}")

        # Check if cached modelled observations exist
        # if self.modelled_obs is None or recalculate:
        if parameter is None or recalculate:
            # Check if observations are present and use these for resampling
            if self.obs is not None:
                self.combine_obs_footprint(
                    resample_to, platform=platform, recalculate=recalculate, cache=True
                )
            else:
                self.scenario = self._check_footprint_resample(resample_to)
        else:
            if self.obs is not None:
                # Check previous resample_to input for cached data
                # prev_resample_to = self.modelled_obs.attrs.get("resample_to")
                prev_resample_to = parameter.attrs.get("resample_to")

                # Check if this previous resample period matches input value
                # - if not (or explicit recalculation requested), recreate scenario
                # - if so return cached modelled observations
                if prev_resample_to != resample_to or recalculate:
                    self.combine_obs_footprint(resample_to, platform=platform, cache=True)
                else:
                    # return self.modelled_obs
                    return False
            elif recalculate:
                # Recalculate based on footprint data if obs not present
                self.scenario = self._check_footprint_resample(resample_to)

            # TODO: Add check for matching sources and recalculate otherwise
            else:
                # Return cached modelled observations if explicit recalculation not requested
                # return self.modelled_obs
                return False

        return True

    def calc_modelled_obs(
        self,
        sources: Optional[Union[str, List]] = None,
        resample_to: str = "coarsest",
        platform: Optional[str] = None,
        cache: bool = True,
        recalculate: bool = False,
    ) -> DataArray:
        """
        Calculate the modelled observation points based on site footprint and fluxes.

        The time points returned are dependent on the resample_to option chosen.
        If obs data is also linked to the ModelScenario instance, this will be used
        to derive the time points where appropriate.

        Args:
            sources: Sources to use for flux. All will be used and stacked if not specified.
            resample_to: Resample option to use for averaging:
                          - either one of ["coarsest", "obs", "footprint"] to match to the datasets
                          - or using a valid pandas resample period e.g. "2H".
                         Default = "coarsest".
            platform: Observation platform used to decide whether to resample e.g. "site", "satellite".
            cache: Cache this data after calculation. Default = True.
            recalculate: Make sure to recalculate this data rather than return from cache. Default = False.

        Returns:
            xarray.DataArray: Modelled observation values along the time axis

            If cache is True:
                This data will also be cached as the ModelScenario.modelled_obs attribute.
                The associated scenario data will be cached as the ModelScenario.scenario attribute.
        """

        self._check_data_is_present(need=["footprint", "fluxes"])

        param_calculate = self._param_setup(
            param="modelled_obs", resample_to=resample_to, platform=platform, recalculate=recalculate
        )

        if not param_calculate:
            modelled_obs = cast(DataArray, self.modelled_obs)
            return modelled_obs

        # Check species and use high time resolution steps if this is carbon dioxide
        if self.species == "co2":
            modelled_obs = self._calc_modelled_obs_HiTRes(
                sources=sources, output_TS=True, output_fpXflux=False
            )
            name = "mf_mod_high_res"
        else:
            modelled_obs = self._calc_modelled_obs_integrated(
                sources=sources, output_TS=True, output_fpXflux=False
            )
            name = "mf_mod"

        modelled_obs.attrs["resample_to"] = resample_to
        modelled_obs = modelled_obs.rename(name)

        # Cache output from calculations
        if cache:
            print("Caching calculated data")
            self.modelled_obs = modelled_obs
            # self.scenario[name] = modelled_obs
        else:
            self.modelled_obs = None  # Make sure this is reset and not cached
            self.scenario = None  # Reset this to None after calculation completed

        return modelled_obs

    def _calc_modelled_obs_integrated(
        self,
        sources: Optional[Union[str, List]] = None,
        output_TS: bool = True,
        output_fpXflux: bool = False,
    ) -> Any:
        """
        Calculate modelled mole fraction timeseries using integrated footprints data.

        Args:
            sources : Flux sources to use for the calculation. By default this will use all available sources.
            output_TS : Whether to output the modelled mole fraction timeseries DataArray.
                       Default = True
            output_fpXflux : Whether to output the modelled flux map DataArray used to create
                            the timeseries. Default = False

        Returns:
            DataArray / DataArray :
                Modelled mole fraction timeseries, dimensions = (time)
                Modelled flux map, dimensions = (lat, lon, time)

            If one of output_TS and output_fpXflux are True:
                DataArray is returned for the respective output

            If both output_TS and output_fpXflux are both True:
                Both DataArrays are returned.
        """

        if self.scenario is None:
            raise ValueError("Combined data must have been defined before calling this function.")

        scenario = self.scenario
        flux = self.combine_flux_sources(sources)
        scenario, flux = match_dataset_dims([scenario, flux], dims=["lat", "lon"])

        flux = flux.reindex_like(scenario, "ffill")
        flux_modelled: DataArray = scenario["fp"] * flux["flux"]
        timeseries: DataArray = flux_modelled.sum(["lat", "lon"])

        if output_TS and output_fpXflux:
            return timeseries, flux_modelled
        elif output_TS:
            return timeseries
        elif output_fpXflux:
            return flux_modelled

    def _calc_modelled_obs_HiTRes(
        self,
        sources: Optional[Union[str, List]] = None,
        averaging: Optional[str] = None,
        output_TS: bool = True,
        output_fpXflux: bool = False,
    ) -> Any:
        """
        Calculate modelled mole fraction timeseries using high time resolution
        footprints data and emissions data. This is appropriate for time variable
        species reliant on high time resolution footprints such as carbon dioxide (co2).

        Args:
            sources : Flux sources to use for the calculation. By default this will use all available sources.
            averaging : Time resolution to use to average the time dimension. Default = None
            output_TS : Whether to output the modelled mole fraction timeseries DataArray.
                       Default = True
            output_fpXflux : Whether to output the modelled flux map DataArray used to create
                            the timeseries. Default = False

        Returns:
            DataArray / DataArray :
                Modelled mole fraction timeseries, dimensions = (time)
                Modelled flux map, dimensions = (lat, lon, time)

            If one of output_TS and output_fpXflux are True:
                DataArray is returned for the respective output

            If both output_TS and output_fpXflux are both True:
                Both DataArrays are returned.

        TODO: Low frequency flux values may need to be selected from the month before
        (currently selecting the same month).
        TODO: Indexing for low frequency flux should be checked to make sure this
        allows for crossing over the end of the year.
        TODO: Currently using pure dask arrays (based on Hannah's original code)
        but would be good to update this to add more pre-indexing using xarray
        and/or use dask as part of datasets.
        TODO: May want to update this to not rely on indexing when selecting
        the appropriate flux values. At the moment this solution has been chosen
        because selecting on a dimension, rather than indexing, can be *very* slow
        depending on the operations performed beforehand on the Dataset (e.g.
        resample and reindex)
        TODO: This code currently resamples the frequency to be regular. This will
        have no effect if the time frequency was already regular but this may
        not be what we want and may want to add extra code to remove any NaNs, if
        they are introduced or to find a way to remove this requirement.
        TODO: mypy having trouble with different types options and incompatible types,
        included as Any for now.
        """
        from math import gcd

        import dask.array as da  # type: ignore
        from pandas import date_range
        from tqdm import tqdm

        # TODO: Need to work out how this fits in with high time resolution method
        # Do we need to flag low resolution to use a different method? natural / anthro for example

        if self.scenario is None:
            raise ValueError("Combined data must have been defined before calling this function.")

        fp_HiTRes = self.scenario.fp_HiTRes
        flux_ds = self.combine_flux_sources(sources)
        fp_HiTRes, flux_ds = match_dataset_dims([fp_HiTRes, flux_ds], dims=["lat", "lon"])

        # Make sure any NaN values are set to zero as this is a multiplicative and summing operation
        fp_HiTRes = fp_HiTRes.fillna(0.0)
        flux_ds["flux"] = flux_ds["flux"].fillna(0.0)

        # Calculate time resolution for both the flux and footprints data
        nanosecond_to_hour = 1 / (1e9 * 60.0 * 60.0)
        flux_res_H = int(flux_ds.time.diff(dim="time").values.mean() * nanosecond_to_hour)
        fp_res_time_H = int(fp_HiTRes.time.diff(dim="time").values.mean() * nanosecond_to_hour)

        fp_res_Hback_H = int(fp_HiTRes["H_back"].diff(dim="H_back").values.mean())

        # Define resolution on time dimension in number in hours
        if averaging:
            try:
                time_res_H = int(averaging)
                time_resolution = f"{time_res_H}H"
            except (ValueError, TypeError):
                time_res_H = int(averaging[0])
                time_resolution = averaging
        else:
            # If not specified derive from time from combined dataset
            time_res_H = fp_res_time_H
            time_resolution = f"{time_res_H}H"

        # Resample fp timeseries to match time resolution
        if fp_res_time_H != time_res_H:
            fp_HiTRes = fp_HiTRes.resample(time=time_resolution).ffill()

        # Define resolution on high frequency dimension in number of hours
        # At the moment this is matched to the Hback dimension
        time_hf_res_H = fp_res_Hback_H

        # Only allow for high frequency resolution < 24 hours
        if time_hf_res_H > 24:
            raise ValueError(f"High frequency resolution must be <= 24 hours. Current: {time_hf_res_H}H")
        elif 24 % time_hf_res_H != 0 or 24 % time_hf_res_H != 0.0:
            raise ValueError(
                f"High frequency resolution must exactly divide into 24 hours. Current: {time_hf_res_H}H"
            )

        # Find the greatest common denominator between time and high frequency resolutions.
        # This is needed to make sure suitable flux frequency is used to allow for indexing.
        # e.g. time: 1H; hf (high frequency): 2H, highest_res_H would be 1H
        # e.g. time: 2H; hf (high frequency): 3H, highest_res_H would be 1H
        highest_res_H = gcd(time_res_H, time_hf_res_H)
        highest_resolution = f"{highest_res_H}H"

        # create time array to loop through, with the required resolution
        # fp_HiTRes.time is the release time of particles into the model
        time_array = fp_HiTRes["time"]
        lat = fp_HiTRes["lat"]
        lon = fp_HiTRes["lon"]
        hback = fp_HiTRes["H_back"]

        ntime = len(time_array)
        nlat = len(lat)
        nlon = len(lon)
        # nh_back = len(hback)

        # Define maximum hour back
        max_h_back = int(hback.values[-1])

        # Define full range of dates to select from the flux input
        date_start = time_array[0]
        date_start_back = date_start - np.timedelta64(max_h_back, "h")
        date_end = time_array[-1] + np.timedelta64(1, "s")

        start = {
            dd: getattr(np.datetime64(time_array[0].values, "h").astype(object), dd)
            for dd in ["month", "year"]
        }

        # Create times for matching to the flux
        full_dates = date_range(
            date_start_back.values, date_end.values, freq=highest_resolution, closed="left"
        ).to_numpy()

        # Create low frequency flux data (monthly)
        flux_ds_low_freq = flux_ds.resample({"time": "1MS"}).mean().sel(time=slice(date_start_back, date_end))
        flux_ds_low_freq = flux_ds_low_freq.transpose(*("lat", "lon", "time"))

        # Select and align high frequency flux data
        flux_ds_high_freq = flux_ds.sel(time=slice(date_start_back, date_end))
        if flux_res_H <= 24:
            base = (
                date_start_back.dt.hour.data
                + date_start_back.dt.minute.data / 60.0
                + date_start_back.dt.second.data / 3600.0
            )
            if flux_res_H <= highest_res_H:
                # Downsample flux to match to footprints frequency
                flux_ds_high_freq = flux_ds_high_freq.resample({"time": highest_resolution}, base=base).mean()
            elif flux_res_H > highest_res_H:
                # Upsample flux to match footprints frequency and forward fill
                flux_ds_high_freq = flux_ds_high_freq.resample(
                    {"time": highest_resolution}, base=base
                ).ffill()
            # Reindex to match to correct values
            flux_ds_high_freq = flux_ds_high_freq.reindex({"time": full_dates}, method="ffill")
        elif flux_res_H > 24:
            # If flux is not high frequency use the monthly averages instead.
            flux_ds_high_freq = flux_ds_low_freq

        # TODO: Add check to make sure time values are exactly aligned based on date range

        # Make sure the dimensions match the expected order for indexing
        fp_HiTRes = fp_HiTRes.transpose(*("lat", "lon", "time", "H_back"))
        flux_ds_high_freq = flux_ds_high_freq.transpose(*("lat", "lon", "time"))

        # Extract footprints array to use in numba loop
        fp_HiTRes = da.array(fp_HiTRes)

        # Set up a numpy array to calculate the product of the footprints (H matrix) with the fluxes
        if output_fpXflux:
            fpXflux = da.zeros((nlat, nlon, ntime))

        if output_TS:
            timeseries = da.zeros(ntime)

        # Iterate through the time coord to get the total mf at each time step using the H back coord
        # at each release time we disaggregate the particles backwards over the previous 24hrs
        # The final value then contains the 29-day integrated residual footprints
        print("Calculating modelled timeseries comparison:")
        iters = tqdm(time_array.values)
        for tt, time in enumerate(iters):

            # Get correct index for low resolution data based on start and current date
            current = {dd: getattr(np.datetime64(time, "h").astype(object), dd) for dd in ["month", "year"]}
            tt_low = current["month"] - start["month"] + 12 * (current["year"] - start["year"])

            # get 4 dimensional chunk of high time res footprints for this timestep
            # units : mol/mol/mol/m2/s
            # reverse the time coordinate to be chronological
            fp_time = fp_HiTRes[:, :, tt, ::-1]

            fp_high_freq = fp_time[:, :, 1:]
            fp_residual = fp_time[:, :, 0:1]  # First element (reversed) contains residual footprints

            # Extract flux data from dataset
            flux_high_freq = flux_ds_high_freq.flux
            flux_low_freq = flux_ds_low_freq.flux

            # Define high and low frequency fluxes based on inputs
            # Allow for variable frequency within 24 hours
            flux_low_freq = flux_low_freq[:, :, tt_low : tt_low + 1]
            if flux_res_H <= 24:
                # Define indices to correctly select matching date range from flux data
                # This will depend on the various frequencies of the inputs
                # At present, highest_res_H matches the flux frequency
                tt_start = tt * int(time_res_H / highest_res_H) + 1
                tt_end = tt_start + int(max_h_back / highest_res_H)
                selection = int(time_hf_res_H / highest_res_H)

                # Extract matching time range from whole flux array
                flux_high_freq = flux_high_freq[..., tt_start:tt_end]
                if selection > 1:
                    # If flux frequency does not match to the high frequency (hf, H_back)
                    # dimension, select entries which do. Reversed to make sure
                    # entries matching to the correct times are selected
                    flux_high_freq = flux_high_freq[..., ::-selection]
                    flux_high_freq = flux_high_freq[..., ::-1]
            else:
                flux_high_freq = flux_high_freq[:, :, tt_low : tt_low + 1]

            # convert to array to use in numba loop
            flux_high_freq = da.array(flux_high_freq)
            flux_low_freq = da.array(flux_low_freq)

            # Multiply the HiTRes footprints with the HiTRes emissions to give mf
            # Multiply residual footprints by low frequency emissions data to give residual mf
            # flux units : mol/m2/s;       fp units : mol/mol/mol/m2/s
            # --> mol/mol/mol/m2/s * mol/m2/s === mol / mol
            fpXflux_time = flux_high_freq * fp_high_freq
            fpXflux_residual = flux_low_freq * fp_residual

            # append the residual emissions
            fpXflux_time = np.dstack((fpXflux_time, fpXflux_residual))

            if output_fpXflux:
                # Sum over time (H back) to give the total mf at this timestep
                fpXflux[:, :, tt] = fpXflux_time.sum(axis=2)

            if output_TS:
                # work out timeseries by summing over lat, lon (24 hrs)
                timeseries[tt] = fpXflux_time.sum()

        if output_fpXflux:
            fpXflux = DataArray(
                fpXflux,
                dims=("lat", "lon", "time"),
                coords={"lat": lat, "lon": lon, "time": time_array},
            )

        if output_TS:
            timeseries = DataArray(timeseries, dims=("time"), coords={"time": time_array})

        if output_fpXflux and output_TS:
            timeseries.compute()
            fpXflux.compute()
            return timeseries, fpXflux
        elif output_fpXflux:
            fpXflux.compute()
            return fpXflux
        elif output_TS:
            timeseries.compute()
            return timeseries

        return None

    def calc_modelled_baseline(
        self,
        resample_to: str = "coarsest",
        platform: Optional[str] = None,
        cache: bool = True,
        recalculate: bool = False,
    ) -> DataArray:
        """
        Calculate the modelled baseline points based on site footprint and boundary conditions.
        Boundary conditions are multipled by any loss (exp(-t/lifetime)) for the species.

        The time points returned are dependent on the resample_to option chosen.
        If obs data is also linked to the ModelScenario instance, this will be used
        to derive the time points where appropriate.

        Args:
            resample_to: Resample option to use for averaging:
                          - either one of ["coarsest", "obs", "footprint"] to match to the datasets
                          - or using a valid pandas resample period e.g. "2H".
                         Default = "coarsest".
            platform: Observation platform used to decide whether to resample e.g. "site", "satellite".
            cache: Cache this data after calculation. Default = True.
            recalculate: Make sure to recalculate this data rather than return from cache. Default = False.

        Returns:
            xarray.DataArray: Modelled baselined values along the time axis

            If cache is True:
                This data will also be cached as the ModelScenario.modelled_baseline attribute.
                The associated scenario data will be cached as the ModelScenario.scenario attribute.
        """
        from openghg.util import check_lifetime_monthly, species_lifetime, time_offset

        self._check_data_is_present(need=["footprint", "bc"])
        bc = cast(BoundaryConditionsData, self.bc)

        param_calculate = self._param_setup(
            param="modelled_baseline", resample_to=resample_to, platform=platform, recalculate=recalculate
        )

        if not param_calculate:
            modelled_baseline = cast(DataArray, self.modelled_baseline)
            return modelled_baseline

        scenario = cast(Dataset, self.scenario)
        bc_data = bc.data

        bc_data = bc_data.reindex_like(scenario, "ffill")

        lifetime_value = species_lifetime(self.species)
        check_monthly = check_lifetime_monthly(lifetime_value)

        if check_monthly:
            lifetime_monthly = cast(Optional[List[str]], lifetime_value)
            lifetime: Optional[str] = None
        else:
            lifetime_monthly = None
            lifetime = cast(Optional[str], lifetime_value)

        if lifetime is not None:
            short_lifetime = True
            lt_time_delta = time_offset(period=lifetime)
            lifetime_hrs: Union[float, np.ndarray] = lt_time_delta.total_seconds() / 3600.0
        elif lifetime_monthly:
            short_lifetime = True
            lifetime_monthly_hrs = []
            for lt in lifetime_monthly:
                lt_time_delta = time_offset(period=lt)
                lt_hrs = lt_time_delta.total_seconds() / 3600.0
                lifetime_monthly_hrs.append(lt_hrs)

            # calculate the lifetime_hrs associated with each time point in scenario data
            # this is because lifetime can be a list of monthly values
            time_month = scenario["time"].dt.month
            lifetime_hrs = np.array([lifetime_monthly_hrs[item - 1] for item in time_month.values])
        else:
            short_lifetime = False

        # Include loss condition if lifetime of species is specified
        if short_lifetime:
            expected_vars = (
                "mean_age_particles_n",
                "mean_age_particles_e",
                "mean_age_particles_s",
                "mean_age_particles_w",
            )
            for var in expected_vars:
                if var not in scenario.data_vars:
                    raise ValueError(
                        f"Unable to calculate baseline for short-lived species {self.species} without species specific footprint."
                    )

            # Ignoring type below -  - problem with xarray patching np.exp to return DataArray rather than ndarray
            loss_n: Union[DataArray, float] = np.exp(-1 * scenario["mean_age_particles_n"] / lifetime_hrs).rename("loss_n")  # type: ignore
            loss_e: Union[DataArray, float] = np.exp(-1 * scenario["mean_age_particles_e"] / lifetime_hrs).rename("loss_e")  # type: ignore
            loss_s: Union[DataArray, float] = np.exp(-1 * scenario["mean_age_particles_s"] / lifetime_hrs).rename("loss_s")  # type: ignore
            loss_w: Union[DataArray, float] = np.exp(-1 * scenario["mean_age_particles_w"] / lifetime_hrs).rename("loss_w")  # type: ignore

        else:

            loss_n = 1.0
            loss_e = 1.0
            loss_s = 1.0
            loss_w = 1.0

        modelled_baseline = (
            (scenario["particle_locations_n"] * bc_data["vmr_n"] * loss_n).sum(["height", "lon"])
            + (scenario["particle_locations_e"] * bc_data["vmr_e"] * loss_e).sum(["height", "lat"])
            + (scenario["particle_locations_s"] * bc_data["vmr_s"] * loss_s).sum(["height", "lon"])
            + (scenario["particle_locations_w"] * bc_data["vmr_w"] * loss_w).sum(["height", "lat"])
        )

        modelled_baseline.attrs["resample_to"] = resample_to
        modelled_baseline = modelled_baseline.rename("bc_mod")

        # Cache output from calculations
        if cache:
            print("Caching calculated data")
            self.modelled_baseline = modelled_baseline
            # self.scenario[name] = modelled_obs
        else:
            self.modelled_baseline = None  # Make sure this is reset and not cached
            self.scenario = None  # Reset this to None after calculation completed

        return modelled_baseline

    # def _calc_modelled_baseline_long_lived():
    #     pass

    # def _calc_modelled_baseline_short_lived():
    #     pass

    def footprints_data_merge(
        self,
        resample_to: str = "coarsest",
        platform: Optional[str] = None,
        calc_timeseries: bool = True,
        sources: Optional[Union[str, List]] = None,
        calc_bc: bool = True,
        cache: bool = True,
        recalculate: bool = False,
    ) -> Dataset:
        """
        Produce combined object containing aligned footprint and observation data.
        Can also include modelled timeseries data derived from flux.

        Args:
            resample_to: Resample option to use for averaging:
                          - either one of ["coarsest", "obs", "footprint"] to match to the datasets
                          - or using a valid pandas resample period e.g. "2H".
                         Default = "coarsest".
            platform: Observation platform used to decide whether to resample.
            calc_timeseries: Calculate modelled timeseries based on flux sources.
            sources: Sources to use for flux if calc_timseries is True.
                     All will be used and stacked if not specified.
            calc_baseline: Calculate modelled baseline.
            cache: Cache this data after calculation. Default = True.
            recalculate: Make sure to recalculate this data rather than return from cache. Default = False.

        Returns:
            xarray.Dataset: Combined dataset containing footprint and observation data
        """
        combined_dataset = self.combine_obs_footprint(
            resample_to=resample_to, platform=platform, cache=cache, recalculate=recalculate
        )

        if calc_timeseries:
            modelled_obs = self.calc_modelled_obs(
                resample_to=resample_to,
                sources=sources,
                platform=platform,
                cache=cache,
                recalculate=recalculate,
            )

            name = modelled_obs.name
            combined_dataset = combined_dataset.assign({name: modelled_obs})

        if calc_bc:
            if self.bc is not None:
                modelled_baseline = self.calc_modelled_baseline(
                    resample_to=resample_to,
                    platform=platform,
                    cache=cache,
                    recalculate=recalculate,
                )
                name = modelled_baseline.name
                combined_dataset = combined_dataset.assign({name: modelled_baseline})
            else:
                print(
                    "Unable to calculate baseline data. Add boundary conditions using ModelScenarion.add_bc(...) to do this."
                )

        if cache:
            self.scenario = combined_dataset

        return combined_dataset

    def plot_timeseries(self) -> Any:
        """
        Plot the observation timeseries data.

        Returns:
            Plotly Figure

            Interactive plotly graph created with observations
        """
        self._check_data_is_present(need="obs")
        obs = cast(ObsData, self.obs)

        fig = obs.plot_timeseries()  # Calling method on ObsData class

        return fig

    def plot_comparison(
        self,
        baseline: Optional[str] = "boundary_conditions",
        sources: Optional[Union[str, List]] = None,
        resample_to: str = "coarsest",
        platform: Optional[str] = None,
        cache: bool = True,
        recalculate: bool = False,
    ) -> Any:
        """
        Plot comparison between observation and modelled timeseries data.

        Args:
            baseline: Add baseline to data. One of:
                          - "boundary_conditions" - Uses added boundary conditions to calculate modelled baseline
                          - "percentile" - Calculates the 1% value across the whole time period
                          - None - don't add a baseline and only plot the modelled observations
            sources: Sources to use for flux. All will be used and stacked if not specified.
            resample_to: Resample option to use for averaging:
                          - either one of ["coarsest", "obs", "footprint"] to match to the datasets
                          - or using a valid pandas resample period e.g. "2H".
                         Default = "coarsest".
            platform: Observation platform used to decide whether to resample e.g. "site", "satellite".
            cache: Cache this data after calculation. Default = True.
            recalculate: Make sure to recalculate this data rather than return from cache. Default = False.

        Returns:
            Plotly Figure

            Interactive plotly graph created with observation and modelled observation data.
        """
        # Only import plotly when we need to - not needed if not plotting.
        import plotly.graph_objects as go

        self._check_data_is_present(need=["obs", "footprint", "flux"])
        obs = cast(ObsData, self.obs)

        fig = obs.plot_timeseries()

        modelled_obs = self.calc_modelled_obs(
            sources=sources, resample_to=resample_to, platform=platform, cache=cache, recalculate=recalculate
        )
        x_data = modelled_obs["time"]
        y_data = modelled_obs.data

        species = self.species
        if sources is None:
            sources = self.flux_sources
        elif isinstance(sources, str):
            sources = [sources]

        if sources is not None:
            source_str = ", ".join(sources)
            label = f"Modelled {species.upper()}: {source_str}"
        else:
            label = f"Modelled {species.upper()}"

        if baseline == "boundary_conditions":
            if self.bc is not None:
                modelled_baseline = self.calc_modelled_baseline(
                    resample_to=resample_to, platform=platform, cache=cache, recalculate=recalculate
                )
                y_baseline = modelled_baseline.data
                y_data = y_data + y_baseline
            else:
                print("Unable to calculate baseline from boundary conditions")
        elif baseline == "percentile":
            mf = obs.data["mf"]
            y_baseline = mf.quantile(1.0, dim="time").values
            y_data = y_data + y_baseline

        fig.add_trace(go.Scatter(x=x_data, y=y_data, mode="lines", name=label))

        return fig


def _indexes_match(dataset_A: Dataset, dataset_B: Dataset) -> bool:
    """
    Check if two datasets need to be reindexed_like for combine_datasets

    Args:
        dataset_A: First dataset to check
        dataset_B: Second dataset to check
    Returns:
        bool: True if indexes match, else False
    """
    common_indices = (key for key in dataset_A.indexes.keys() if key in dataset_B.indexes.keys())

    for index in common_indices:
        if not len(dataset_A.indexes[index]) == len(dataset_B.indexes[index]):
            return False

        # Check number of values that are not close (testing for equality with floating point)
        if index == "time":
            # For time override the default to have ~ second precision
            rtol = 1e-10
        else:
            rtol = 1e-5

        index_diff = np.sum(
            ~np.isclose(
                dataset_A.indexes[index].values.astype(float),
                dataset_B.indexes[index].values.astype(float),
                rtol=rtol,
            )
        )

        if not index_diff == 0:
            return False

    return True


def combine_datasets(
    dataset_A: Dataset, dataset_B: Dataset, method: methodType = "ffill", tolerance: Optional[float] = None
) -> Dataset:
    """
    Merges two datasets and re-indexes to the first dataset.

    If "fp" variable is found within the combined dataset,
    the "time" values where the "lat", "lon" dimensions didn't match are removed.

    Args:
        dataset_A: First dataset to merge
        dataset_B: Second dataset to merge
        method: One of None, nearest, ffill, bfill.
                See xarray.DataArray.reindex_like for list of options and meaning.
                Defaults to ffill (forward fill)
        tolerance: Maximum allowed tolerance between matches.
    Returns:
        xarray.Dataset: Combined dataset indexed to dataset_A
    """
    if _indexes_match(dataset_A, dataset_B):
        dataset_B_temp = dataset_B
    else:
        dataset_B_temp = dataset_B.reindex_like(dataset_A, method=method, tolerance=tolerance)

    merged_ds = dataset_A.merge(dataset_B_temp)

    if "fp" in merged_ds:
        if all(k in merged_ds.fp.dims for k in ("lat", "long")):
            flag = np.where(np.isfinite(merged_ds.fp.mean(dim=["lat", "lon"]).values))
            merged_ds = merged_ds[dict(time=flag[0])]

    return merged_ds


def match_dataset_dims(
    datasets: Sequence[Dataset],
    dims: Union[str, Sequence] = [],
    method: methodType = "nearest",
    tolerance: Union[float, Dict[str, float]] = 1e-5,
) -> List[Dataset]:
    """
    Aligns datasets to the selected dimensions within a tolerance.
    All datasets will be aligned to the first dataset within the list.

    Args:
        datasets: List of xarray Datasets. Expect datasets to contain the same dimensions.
        dims: Dimensions match between datasets. Can use keyword "all" to match every dimension.
        method : Method to use for indexing. Should be one of: ("nearest", "ffill", "bfill")
        tolerance: Tolerance value to use when matching coordinate values.
                   This can be a single value for all dimensions or a dictionary of values to use.

    Returns:
        List (xarray.Dataset) : Datasets aligned along the matching dimensions.

    TODO: Check if this supercedes or replicates _indexes_match() function too closely?
    """

    # Nothing to be done if only one (or less) datasets are passed
    if len(datasets) <= 1:
        return list(datasets)

    ds0 = datasets[0]

    if isinstance(dims, str):
        if dims == "all":
            dims = list(ds0.dims)
        else:
            dims = [dims]

    # Extract coordinate values for the first dataset in the list
    ds0 = datasets[0]
    indexers = {dim: ds0[dim] for dim in dims}

    if isinstance(tolerance, float):
        tolerance = {dim: tolerance for dim in dims}

    # Align datasets along selected dimensions (if not already identical)
    datasets_aligned = [ds0]
    for ds in datasets[1:]:
        for dim, compare_coord in indexers.items():
            try:
                coord = ds[dim]
            except KeyError:
                raise ValueError(f"Dataset missing dimension: {dim}")
            else:
                if not coord.equals(compare_coord):
                    ds = ds.reindex({dim: compare_coord}, method=method, tolerance=tolerance[dim])

        datasets_aligned.append(ds)

    return datasets_aligned


# ResType = Union[np.timedelta64, float, np.floating, np.integer]


def calc_dim_resolution(dataset: Dataset, dim: str = "time") -> Any:
    """
    Calculates the average frequency along a given dimension.

    Args:
        dataset : Dataset. Must contain the specified dimension
        dim : Dimension name

    Returns:
        np.timedelta64 / np.float / np.int : Resolution with input dtype

        NaT : If unsuccessful and input dtype is np.timedelta64
        NaN : If unsuccessful for all other dtypes.
    """
    try:
        resolution = dataset[dim].diff(dim=dim).mean().item()
    except ValueError:
        if np.issubdtype(dataset[dim].dtype, np.datetime64):
            resolution = np.timedelta64("NaT")
        else:
            resolution = np.NaN
    else:
        if np.issubdtype(dataset[dim].dtype, np.datetime64):
            # Extract units from original datetime string and use to recreate timedelta64
            unit = dataset[dim].dtype.name.lstrip("timedelta64").lstrip("[").rstrip("]")
            resolution = np.timedelta64(resolution, unit)

    return resolution


def stack_datasets(datasets: Sequence[Dataset], dim: str = "time", method: methodType = "ffill") -> Dataset:
    """
    Stacks multiple datasets based on the input dimension. By default this is time
    and this will be aligned to the highest resolution / frequency
    (smallest difference betweeen coordinate values).

    At the moment, the two datasets must have identical coordinate values for all
    other dimensions and variable names for these to be stacked.

    Args:
        datasets : Sequence of input datasets
        dim : Name of dimension to stack along. Default = "time"
        method: Method to use when aligning the datasets. Default = "ffill"

    Returns:
        Dataset : Stacked dataset

    TODO: Could update this to only allow DataArrays to be included to reduce the phase
    space here.
    """

    if len(datasets) == 1:
        dataset = datasets[0]
        return dataset

    data_frequency = [calc_dim_resolution(ds, dim) for ds in datasets]
    index_highest_freq = min(range(len(data_frequency)), key=data_frequency.__getitem__)
    data_highest_freq = datasets[index_highest_freq]
    coords_to_match = data_highest_freq[dim]

    for i, data in enumerate(datasets):
        data_match = data.reindex({dim: coords_to_match}, method=method)
        if i == 0:
            data_stacked = data_match
            data_stacked.attrs = {}
        else:
            data_stacked += data_match

    return data_stacked


# def footprints_data_merge(data: Union[dict, ObsData],
#                           domain: str,
#                           met_model: Optional[str] = None,
#                           load_flux: bool = True,
#                           load_bc: bool = True,
#                           calc_timeseries: bool = True,
#                           calc_bc: bool = True,
#                           HiTRes: bool = False,
#                           site_modifier: Dict[str, str] = {},
#                           height: Optional[str] = None,
#                           emissions_name: Optional[str] = None,
#                           fp_directory: Optional[Union[Path, str]] = None,
#                           flux_directory: Optional[Union[Path, str]] = None,
#                           bc_directory: Optional[Union[Path, str]] = None,
#                           resample_to_data: bool = False,
#                           species_footprint: Optional[str] = None,
#                           chunks: bool = False,
#                           verbose: bool = True,
#                           ) -> Any:
#     """
#     This will be a wrapper for footprints_data_merge function from acrg_name.name file written
#     """

#     # Write this a wrapper for footprints_data_merge function from acrg_name.name file

#     print("The footprint_data_merge() wrapper function will be deprecated.")
#     print("Please use the ModelScenario class to set up your data")
#     print("Then call the model.footprints_data_merge() method e.g.")
#     print(" model = ModelScenario(site, species, inlet, network, domain, ...)")
#     print(" combined_data = model.footprints_data_merge()")

#     # resample_to_data --> resample_to
#     # HiTRes --> linked to species as co2
#     # directories --> could link to adding new data to object store?
#     # height --> inlet
#     # species_footprint --> links to overall species?
#     # site_modifier --> link to footprint name - different site name for footprint?


# Blueprint from Issue #42 created for this

# class LPDM():

#     def __init__(species, domain, flux, bc):
#         self.species=species
#         self.domain=domain
#         #etc...

#         self.flux=get_flux(species, domain, flux)
#         self.bc=get_bc(species, domain, bc)

#         # dictionary of obs datasets + footprints?
#         self.obs = {}
#         self.footprints = {}

#     def obs_get(self, site, average=None):
#         self.obs["site"]=get_observations(self.species, site)

#     def obs_footprints_merge(self.obs, site, etc.):
#         obs_resampled, fp_resampled = footprints_data_merge(self.obs, site, kwargs_see_above)
#         self.obs["site"] = obs_resampled
#         self.footprints["site"] = fp_resampled

#     def model_mf(self, site):
#         # this is obviously not to be taken literally:
#         return self.footprints["site"] * self.flux + self.bc


# if __name__ == "__main__":
#     import LPDM # Need a better class name?

#     mod = LPDM(species="CH4", domain="EUROPE", flux=["ch4_source1", "ch4_source2", "ch4_source3"],
#                         bc="CAMS")

#     for site in sites:
#         mod.obs_get(site, average="1H") # Need to be careful about averaging here and/or in fp/data merge step
#         mod.obs_footprints_merge(site, model="NAME-UKV", average="1D") # 1D average of combined fp/data dataset

#         # Optional
#         mod.get_site_met(site, met="ECMWF")

#     # output some obs
#     ds = mod.obs_out(site)
#     mod.obs_plot(site)

#     # plot some model output
#     mod.footprint_plot(site)

#     # predicted mole fraction
#     mf_mod = mod.model_mf(site, source="ch4_source1")



from pathlib import Path
from typing import Dict, Optional, Union


def parse_btt(
    data_filepath: Union[str, Path],
    site: Optional[str] = "BTT",
    network: Optional[str] = "LGHG",
    inlet: Optional[str] = None,
    instrument: Optional[str] = None,
) -> Dict:
    """Reads NPL data files and returns the UUIDS of the Datasources
    the processed data has been assigned to

    Args:
        data_filepath: Path of file to load
        site: Site name
    Returns:
        dict: Dictionary of gas data
    """
    from numpy import nan as np_nan
    from openghg.standardise.meta import assign_attributes
    from openghg.util import clean_string, load_json
    from pandas import Timestamp, isnull, read_csv, to_timedelta

    # TODO: Decide what to do about inputs which aren't use anywhere
    # at present - inlet, instrument, sampling_period, measurement_type

    data_filepath = Path(data_filepath)

    site = "BTT"

    # Rename these columns
    rename_dict = {"co2.cal": "CO2", "ch4.cal.ppb": "CH4"}
    # We only want these species
    species_extract = ["CO2", "CH4"]
    # Take std-dev measurements from these columns for these species
    species_sd = {"CO2": "co2.sd.ppm", "CH4": "ch4.sd.ppb"}

    site_data = load_json(filename="acrg_site_info.json")
    site_info = site_data[site][network]

    param_data = load_json(filename="attributes.json")
    network_params = param_data["BTT"]
    site_attributes = network_params["global_attributes"]

    sampling_period = int(network_params["sampling_period"])
    sampling_period_seconds = str(sampling_period) + "s"

    metadata = {}
    metadata["site"] = site
    metadata["inlet"] = network_params["inlet"]
    metadata["instrument"] = network_params["instrument"]
    metadata["sampling_period"] = str(sampling_period)
    metadata["station_longitude"] = site_info["longitude"]
    metadata["station_latitude"] = site_info["latitude"]
    metadata["station_long_name"] = site_info["long_name"]

    attributes = network_params["global_attributes"]
    attributes["inlet_height_magl"] = network_params["inlet"].strip("m")
    attributes.update(metadata)

    data = read_csv(data_filepath)
    data["time"] = Timestamp("2019-01-01 00:00") + to_timedelta(data["DOY"] - 1, unit="D")
    data["time"] = data["time"].dt.round(sampling_period_seconds)
    data = data[~isnull(data.time)]

    data = data.rename(columns=rename_dict)
    data = data.set_index("time")

    gas_data = {}
    for species in species_extract:
        processed_data = data.loc[:, [species]].sort_index()
        # Create a variability column
        species_stddev_label = species_sd[species]
        processed_data[species][f"{species} variability"] = data[species_stddev_label]

        # Replace any values below zero with NaNs
        processed_data[processed_data < 0] = np_nan
        # Drop NaNs
        processed_data = processed_data.dropna()
        # Convert to a Dataset
        processed_data = processed_data.to_xarray()

        species_attributes = attributes.copy()
        species_attributes["species"] = clean_string(species)

        species_metadata = metadata.copy()
        species_metadata["species"] = clean_string(species)

        gas_data[species] = {
            "metadata": species_metadata,
            "data": processed_data,
            "attributes": site_attributes,
        }

    gas_data = assign_attributes(data=gas_data, site=site, network=network)

    return gas_data



"""
This is used as a base for the other dataclasses and shouldn't be used directly.
"""
from dataclasses import dataclass
from typing import Dict

from xarray import Dataset


@dataclass(frozen=True)
class _BaseData:
    data: Dataset
    metadata: Dict

    def __bool__(self) -> bool:
        return bool(self.data)

    def __str__(self) -> str:
        return f"Data: {self.data}\nMetadata : {self.metadata}"



from pathlib import Path
from typing import Dict, Optional, Union


def parse_eurocom(
    data_filepath: Union[str, Path],
    site: str,
    sampling_period: str,
    network: Optional[str] = None,
    inlet: Optional[str] = None,
    instrument: Optional[str] = None,
) -> Dict:
    """Parses EUROCOM data files into a format expected by OpenGHG

    Args:
        data_filepath: Path of file to read
        site: Site code
        sampling_period: Sampling period in seconds
        network: Network name
        Inlet: Inlet height in metres
        Instrument: Instrument name
    Returns:
        dict: Dictionary of measurement data
    """
    from openghg.standardise.meta import assign_attributes, get_attributes
    from openghg.util import load_json, read_header
    from pandas import Timestamp, read_csv

    data_filepath = Path(data_filepath)

    if site is None:
        site = data_filepath.stem.split("_")[0]

    if sampling_period is None:
        sampling_period = "NOT_SET"

    data_filepath = Path(data_filepath)

    filename = data_filepath.name
    inlet_height = filename.split("_")[1]

    if "m" not in inlet_height:
        inlet_height = "NA"

    # This dictionary is used to store the gas data and its associated metadata
    combined_data = {}

    # Read the header as lines starting with #
    header = read_header(data_filepath, comment_char="#")
    n_skip = len(header) - 1
    species = "co2"

    def date_parser(year: str, month: str, day: str, hour: str, minute: str) -> Timestamp:
        return Timestamp(year=year, month=month, day=day, hour=hour, minute=minute)

    datetime_columns = {"time": ["Year", "Month", "Day", "Hour", "Minute"]}
    use_cols = [
        "Day",
        "Month",
        "Year",
        "Hour",
        "Minute",
        str(species.lower()),
        "SamplingHeight",
        "Stdev",
        "NbPoints",
    ]

    dtypes = {
        "Day": int,
        "Month": int,
        "Year": int,
        "Hour": int,
        "Minute": int,
        species.lower(): float,
        "Stdev": float,
        "SamplingHeight": float,
        "NbPoints": int,
    }

    data = read_csv(
        data_filepath,
        skiprows=n_skip,
        parse_dates=datetime_columns,
        date_parser=date_parser,
        index_col="time",
        sep=";",
        usecols=use_cols,
        dtype=dtypes,
        na_values="-999.99",
    )

    data = data[data[species.lower()] >= 0.0]
    data = data.dropna(axis="rows", how="any")
    # Drop duplicate indices
    data = data.loc[~data.index.duplicated(keep="first")]
    # Convert to xarray Dataset
    data = data.to_xarray()

    attributes_data = load_json(filename="attributes.json")
    eurocom_attributes = attributes_data["EUROCOM"]
    global_attributes = eurocom_attributes["global_attributes"]

    if inlet_height == "NA":
        try:
            inlet = eurocom_attributes["intake_height"][site]
            global_attributes["inlet_height_m"] = inlet
            calibration_scale = eurocom_attributes["calibration"][site]
        except KeyError:
            calibration_scale = {}
            raise ValueError(f"Unable to find inlet from filename or attributes file for {site}")

    gas_data = get_attributes(
        ds=data,
        species=species,
        site=site,
        global_attributes=global_attributes,
        units="ppm",
    )

    # Create a copy of the metadata dict
    metadata = {}
    metadata["site"] = site
    metadata["species"] = species
    metadata["inlet_height"] = global_attributes["inlet_height_m"]
    metadata["calibration_scale"] = calibration_scale
    metadata["network"] = "EUROCOM"
    metadata["sampling_period"] = str(sampling_period)

    combined_data[species] = {
        "metadata": metadata,
        "data": gas_data,
        "attributes": global_attributes,
    }

    combined_data = assign_attributes(data=combined_data, site=site, sampling_period=sampling_period)

    return combined_data



from ._footprint import footprints_data_merge, single_site_footprint
from ._scenario import (
    ModelScenario,
    calc_dim_resolution,
    combine_datasets,
    match_dataset_dims,
    stack_datasets,
)



from pathlib import Path
from typing import Optional, Union


def parse_url_filename(url: str) -> str:
    """Get the filename from a (messy) URL.

    Args:
        url: URL of file
    Returns:
        str: Filename
    """
    from urllib.parse import urlparse

    return Path(urlparse(url).path).name


def download_data(
    url: str, filepath: Optional[Union[str, Path]] = None, timeout: int = 10
) -> Optional[bytes]:
    """Download data file, with progress bar.

    Based on https://stackoverflow.com/a/63831344/1303032

    Args:
        url: URL of content to download
        filepath: Filepath to write out data
        timeount: Timeout for HTTP request (seconds)
    Returns:
        bytes / None: Bytes if no filepath given
    """
    import functools
    import io
    import shutil
    from urllib.parse import urlparse

    import requests
    from requests.adapters import HTTPAdapter
    from tqdm.auto import tqdm  # type: ignore
    from urllib3.util.retry import Retry  # type: ignore

    retriable_status_codes = [
        requests.codes.internal_server_error,
        requests.codes.bad_gateway,
        requests.codes.service_unavailable,
        requests.codes.gateway_timeout,
        requests.codes.too_many_requests,
        requests.codes.request_timeout,
    ]

    retry_strategy = Retry(
        total=3,
        status_forcelist=retriable_status_codes,
        allowed_methods=["HEAD", "GET", "OPTIONS"],
        backoff_factor=1,
    )  # type: ignore

    adapter = HTTPAdapter(max_retries=retry_strategy)

    http = requests.Session()
    http.mount("https://", adapter)
    http.mount("http://", adapter)

    try:
        r = http.get(url=url, stream=True, allow_redirects=True, timeout=timeout)
    except (requests.exceptions.RequestException, requests.exceptions.ConnectionError) as e:
        print(f"Unable to retrieve data from {url}, error: {str(e)}")
        return None

    filename = Path(urlparse(url).path).name

    if r.status_code != 200:
        print(f"Unable to download {url}, please check URL.")
        return None

    file_size = int(r.headers.get("Content-Length", 0))

    desc = f"Downloading {filename}"
    r.raw.read = functools.partial(r.raw.read, decode_content=True)

    with tqdm.wrapattr(r.raw, "read", total=file_size, desc=desc) as r_raw:
        with io.BytesIO() as buf:
            shutil.copyfileobj(r_raw, buf)

            if filepath is None:
                return buf.getvalue()
            else:
                Path(filepath).write_bytes(buf.getvalue())
                return None



class InvalidSiteError(Exception):
    """Raised if an invalid site is passed"""


class UnknownDataError(Exception):
    """Raised if an unknown data type is passed"""


class FunctionError(Exception):
    """Raised if a serverless function cannot be called correctly"""


class ObjectStoreError(Exception):
    """Raised if an error accessing an object at a key in the object store occurs"""


class DatasourceLookupError(Exception):
    """Raised if Datasource lookup fails"""


class EncodingError(ObjectStoreError):
    pass


class MutexTimeoutError(Exception):
    pass


class RequestBucketError(Exception):
    pass



import json
from io import BytesIO
from typing import Dict, Iterator, List, Optional, Type, TypeVar, Union, cast

from addict import Dict as aDict
from openghg.dataobjects import ObsData
from openghg.store import recombine_datasets
from openghg.util import (
    clean_string,
    create_daterange_str,
    find_daterange_gaps,
    first_last_dates,
    running_on_hub,
    split_daterange_str,
)
from xarray import Dataset, open_dataset

__all__ = ["SearchResults"]

T = TypeVar("T", bound="SearchResults")


class SearchResults:
    """This class is used to return data from the search function. It
    has member functions to retrieve data from the object store.

    Args:
        results: Search results
        ranked_data: True if results are ranked, else False
    """

    def __init__(self, results: Optional[Dict] = None, ranked_data: bool = False):
        self.results = results if results is not None else {}
        self.ranked_data = ranked_data
        self.hub = running_on_hub()

    def __str__(self) -> str:
        if not self.results:
            return "No results"

        print_strs = []
        for site, species in self.results.items():
            if self.ranked_data:
                print_strs.append(
                    f"Site: {site.upper()} \nSpecies found: {', '.join(self.results[site].keys())}"
                )
            else:
                print_strs.append(f"Site: {site.upper()}")
                print_strs.append("---------")
                print_strs.extend([f"{sp} at {', '.join(self.results[site][sp].keys())}" for sp in species])
            print_strs.append("\n")

        return "\n".join(print_strs)

    def __repr__(self) -> str:
        return self.__str__()

    def __bool__(self) -> bool:
        return bool(self.results)

    def __len__(self) -> int:
        return len(self.results)

    def __iter__(self) -> Iterator:
        yield from self.results

    def to_data(self) -> Dict:
        """Convert this object to a dictionary for JSON serialisation

        Returns:
            dict: Dictionary of data
        """
        return {
            "results": self.results,
            "ranked_data": self.ranked_data,
            "cloud": self.hub,
        }

    def to_json(self) -> str:
        """Serialises the object to JSON

        Returns:
            str: JSON str
        """
        return json.dumps(self.to_data())

    @classmethod
    def from_json(cls: Type[T], data: Union[bytes, str]) -> T:
        """Create a SearchResults object from a dictionary

        Args:
            data: Serialised object
        Returns:
            SearchResults: SearchResults object
        """
        loaded = json.loads(data)

        return cls(results=loaded["results"], ranked_data=loaded["ranked_data"])

    def rankings(self) -> Dict:
        if not self.ranked_data:
            print("No rank data")

        rank_result = aDict()

        for site, species_data in self.results.items():
            for species, data in species_data.items():
                rank_result[site][species] = data["rank_metadata"]

        to_return: Dict = rank_result.to_dict()

        return to_return

    def raw(self) -> Dict:
        """Returns the raw results data

        Returns:
            dict: Dictionary of results returned from search function
        """
        return self.results

    def keys(self, site: str, species: str, inlet: Optional[str] = None) -> Optional[List[str]]:
        """Return the data keys for the specified site and species.
        This is intended mainly for use in the search function when filling
        gaps of unranked dateranges.

            Args:
                site: Three letter site code
                species: Species name
                inlet: Inlet height, required for unranked data
            Returns:
                list: List of keys
        """
        site = site.lower()
        species = species.lower()

        if inlet is not None:
            inlet = inlet.lower()

        try:
            if self.ranked_data:
                keys: List = self.results[site][species]["keys"]
            else:
                keys = self.results[site][species][inlet]["keys"]

            return keys
        except KeyError:
            print(f"No keys found for {species} at {site}")
            return None

    def metadata(self, site: str, species: str, inlet: Optional[str] = None) -> Optional[Dict]:
        """Return the metadata for the specified site and species

        Args:
            site: Three letter site code
            species: Species name
            inlet: Inlet height, required for unranked data
        Returns:
            dict: Dictionary of metadata
        """
        site = site.lower()
        species = species.lower()

        if inlet is None and not self.ranked_data:
            raise ValueError("Please pass an inlet height.")

        if inlet is not None:
            inlet = inlet.lower()

        try:
            if self.ranked_data:
                metadata: Dict = self.results[site][species]["metadata"]
            else:
                metadata = self.results[site][species][inlet]["metadata"]
        except KeyError:
            print(f"No metadata found for {species} at {site}")
            return None
        else:
            return metadata

    def retrieve_all(self) -> Union[ObsData, List[ObsData], None]:
        """Retrieve all the data found during the serch

        Returns:
            list: List of ObsData objects
        """
        results = []

        if self.ranked_data:
            # Can we just traverse the dict without looping?
            for site, species_data in self.results.items():
                for species, inlet_data in species_data.items():
                    obsdata = self._create_obsdata(site=site, species=species)
                    results.append(obsdata)
        else:
            # Can we just traverse the dict without looping?
            for site, species_data in self.results.items():
                for species, inlet_data in species_data.items():
                    for inlet in inlet_data:
                        obsdata = self._create_obsdata(site=site, species=species, inlet=inlet)
                        results.append(obsdata)

        if not results:
            return None
        if len(results) == 1:
            return results[0]
        else:
            return results

    def retrieve(
        self,
        site: Optional[str] = None,
        species: Optional[str] = None,
        inlet: Optional[str] = None,
    ) -> Union[ObsData, List[ObsData], None]:
        """Retrieve some or all of the data found in the object store.

        Args:
            site: Three letter site code
            species: Species name
        Returns:
            ObsData or dictionary of ObsData objects
        """
        site = clean_string(site)
        species = clean_string(species)
        inlet = clean_string(inlet)

        results = []
        if not self.ranked_data:
            for _site, site_data in self.results.items():
                if site is not None and _site != site:
                    continue
                for _species, species_data in site_data.items():
                    if species is not None and _species != species:
                        continue
                    for _inlet in species_data:
                        if inlet is not None and _inlet != inlet:
                            continue

                        obsdata = self._create_obsdata(site=_site, species=_species, inlet=_inlet)

                        if obsdata is not None:
                            results.append(obsdata)
        else:
            if inlet is not None:
                from openghg.retrieve import search

                with_inlet = search(site=site, species=species, inlet=inlet)
                # TODO - remove this cast once we always return a SearchResults object from search
                with_inlet = cast(SearchResults, with_inlet)
                inlet_data = with_inlet.retrieve(site=site, species=species, inlet=inlet)
                return inlet_data

            for _site, site_data in self.results.items():
                if site is not None and _site != site:
                    continue
                for _species, species_data in site_data.items():
                    if species is not None and _species != species:
                        continue

                    obsdata = self._create_obsdata(site=_site, species=_species)

                    if obsdata is not None:
                        results.append(obsdata)

        if not results:
            return None
        if len(results) == 1:
            return results[0]
        else:
            return results

    def _create_obsdata(self, site: str, species: str, inlet: Optional[str] = None) -> ObsData:
        """Creates an ObsData object for return to the user

        Args:
            site: Site code
            species: Species name
        Returns:
            ObsData: ObsData object
        """
        from xarray import concat

        try:
            if self.ranked_data:
                specific_source = self.results[site][species]
            else:
                specific_source = self.results[site][species][inlet]
        except KeyError:
            raise ValueError("Error: We can't create an ObsData object using these parameters.")

        data_keys = specific_source["keys"]
        metadata = specific_source["metadata"]

        if not self.ranked_data:
            keys = data_keys["unranked"]
            final_dataset = self._retrieve_dataset(keys, sort=True, elevate_inlet=False)
        else:
            dataset_slices = []

            inlet_ranges = specific_source["rank_metadata"]

            metadata["rank_metadata"] = {}

            ranked_keys = data_keys["ranked"]
            ranked_slices = []

            inlets = set()

            for daterange, keys in ranked_keys.items():
                data_slice = self._retrieve_dataset(keys=keys, sort=True, elevate_inlet=True)

                slice_start, slice_end = split_daterange_str(daterange_str=daterange, date_only=True)

                # We convert to str here as xarray has some weird behaviour that means
                # "2018-01-01" - "2018-06-01"
                # gets treated differently to
                # datetime.date(2018, 1, 1) - datetime.date(2018, 6, 1)
                ranked_slice = data_slice.sel(time=slice(str(slice_start), str(slice_end)))

                if ranked_slice.time.size > 0:
                    inlets.add(inlet_ranges[daterange])
                    ranked_slices.append(ranked_slice)

                ranked_metadata = specific_source["rank_metadata"]
                metadata["rank_metadata"]["ranked"] = ranked_metadata

            dataset_slices.extend(ranked_slices)

            unranked_keys = data_keys["unranked"]

            if unranked_keys:
                unranked_data = self._retrieve_dataset(keys=unranked_keys, sort=True, elevate_inlet=True)

                first_date, last_date = first_last_dates(keys=unranked_keys)

                ranked_dateranges = list(ranked_keys.keys())
                unranked_dateranges = find_daterange_gaps(
                    start_search=first_date,
                    end_search=last_date,
                    dateranges=ranked_dateranges,
                )

                unranked_metadata = {}
                if unranked_dateranges:
                    unranked_slices = []
                    for dr in unranked_dateranges:
                        slice_start, slice_end = split_daterange_str(daterange_str=dr, date_only=True)
                        unranked_slice = unranked_data.sel(time=slice(str(slice_start), str(slice_end)))

                        if unranked_slice.time.size > 0:
                            _inlet = unranked_slice["inlet"].values[0]
                            inlets.add(_inlet)
                            unranked_metadata[dr] = _inlet
                            unranked_slices.append(unranked_slice)

                    dataset_slices.extend(unranked_slices)
                else:
                    daterange_str = create_daterange_str(start=first_date, end=last_date)
                    _inlet = unranked_data["inlet"].values[0]
                    inlets.add(inlet)
                    unranked_metadata[daterange_str] = _inlet

                    dataset_slices.append(unranked_data)

                metadata["rank_metadata"]["unranked"] = unranked_metadata

            final_dataset = concat(objs=dataset_slices, dim="time").sortby("time")

            if len(inlets) == 1:
                inlet_tag = str(inlets.pop())
            else:
                inlet_tag = "multiple"

            # Update the attributes for single / multiple inlet heights
            final_dataset.attrs["inlet"] = inlet_tag

        metadata = specific_source["metadata"]

        return ObsData(data=final_dataset, metadata=metadata)

    def _retrieve_dataset(
        self, keys: List, sort: bool, elevate_inlet: bool = True, attrs_to_check: Optional[Dict] = None
    ) -> Dataset:
        """Retrieves datasets from either cloud or local object store

        Args:
            keys: List of object store keys
            sort: Sort data on recombination
            elevate_inlet: Elevate inlet from attribute to variable
        Returns:
            Dataset:
        """
        from openghg.cloud import call_function

        if self.hub:
            to_post: Dict[str, Union[Dict, List, bool, str]] = {}
            to_post["keys"] = keys
            to_post["sort"] = sort
            to_post["elevate_inlet"] = elevate_inlet
            to_post["function"] = "retrieve"

            if attrs_to_check is not None:
                to_post["attrs_to_check"] = attrs_to_check

            result = call_function(data=to_post)
            binary_netcdf = result["content"]["data"]
            buf = BytesIO(binary_netcdf)
            # TODO - remove this ignore once xarray have updated their type hints
            ds: Dataset = open_dataset(buf).load()  # type: ignore
            return ds
        else:
            return recombine_datasets(
                keys=keys, sort=sort, elevate_inlet=elevate_inlet, attrs_to_check=attrs_to_check
            )



from openghg.util import running_locally

cloud_env = not running_locally()

if cloud_env:
    from ._oci_store import (
        create_bucket,
        create_par,
        delete_object,
        delete_par,
        exists,
        get_all_object_names,
        get_bucket,
        get_object,
        get_object_from_json,
        set_object,
        set_object_from_file,
        set_object_from_json,
        upload,
    )
    from ._par import PAR
else:
    from ._local_store import (
        delete_object,
        exists,
        get_all_object_names,
        get_bucket,
        get_object,
        get_object_from_json,
        get_object_names,
        get_openghg_local_path,
        query_store,
        set_object,
        set_object_from_file,
        set_object_from_json,
        visualise_store,
    )

from ._encoding import bytes_to_string, datetime_to_datetime, get_datetime_now, string_to_bytes



from dataclasses import dataclass

from ._basedata import _BaseData

__all__ = ["FluxData"]


@dataclass(frozen=True)
class FluxData(_BaseData):
    """This class is used to return flux/emissions data from the get_flux function

    Args:
        data: xarray Dataframe
        metadata: Dictionary of metadata including model run parameters
    """

    def __str__(self) -> str:
        return f"Data: {self.data}\nMetadata : {self.metadata}"



""" Utility functions that are used by multiple modules

"""
from collections.abc import Iterable
from typing import Any, Dict, Iterator, Optional, Tuple


def running_in_cloud() -> bool:
    """Are we running in the cloud?

    Checks for the OPENGHG_CLOUD environment variable being set

    Returns:
        bool: True if running in cloud
    """
    from os import environ

    cloud_env = environ.get("OPENGHG_CLOUD", "0")

    return bool(int(cloud_env))


def running_on_hub() -> bool:
    """Are we running on the OpenGHG Hub?

    Checks for the OPENGHG_CLOUD environment variable being set

    Returns:
        bool: True if running in cloud
    """
    from os import environ

    hub_env = environ.get("OPENGHG_HUB", "0")

    return bool(int(hub_env))


def running_locally() -> bool:
    """Are we running OpenGHG locally?

    Returns:
        bool: True if running locally
    """
    return not (running_on_hub() or running_in_cloud())


def unanimous(seq: Dict) -> bool:
    """Checks that all values in an iterable object
    are the same

    Args:
        seq: Iterable object
    Returns
        bool: True if all values are the same

    """
    it = iter(seq.values())
    try:
        first = next(it)
    except StopIteration:
        return True
    else:
        return all(i == first for i in it)


def pairwise(iterable: Iterable) -> Iterator[Tuple[Any, Any]]:
    """Return a zip of an iterable where a is the iterable
    and b is the iterable advanced one step.

    Args:
        iterable: Any iterable type
    Returns:
        tuple: Tuple of iterables
    """
    from itertools import tee

    a, b = tee(iterable)
    next(b, None)

    return zip(a, b)


def site_code_finder(site_name: str) -> Optional[str]:
    """Find the site code for a given site name.

    Args:
        site_name: Site long name
    Returns:
        str or None: Three letter site code if found
    """
    from openghg.util import load_json
    from rapidfuzz import process  # type: ignore

    sites = load_json("site_lookup.json")

    inverted = {s["short_name"]: c for c, s in sites.items()}

    matches = process.extract(query=site_name, choices=inverted.keys())
    highest_score = matches[0][1]

    if highest_score < 90:
        return None

    matched_site = matches[0][0]
    site_code: str = inverted[matched_site]

    return site_code


def find_matching_site(site_name: str, possible_sites: Dict) -> str:
    """Try and find a similar name to site_name in site_list and return a suggestion or
    error string.

    Args:
        site_name: Name of site
        site_list: List of sites to check
    Returns:
        str: Suggestion / error message
    """
    from rapidfuzz import process

    site_list = possible_sites.keys()

    matches = process.extract(site_name, site_list)

    scores = [s for m, s, _ in matches]

    # This seems like a decent cutoff score for a decent find
    cutoff_score = 85

    if scores[0] < cutoff_score:
        return f"No suggestion for {site_name}."
    elif scores[0] > cutoff_score and scores[0] > scores[1]:
        best_match = matches[0][0]
        return f"Did you mean {best_match.upper()}, code: {possible_sites[best_match]} ?"
    elif scores[0] == scores[1]:
        suggestions = [f"{match.title()}, code: {possible_sites[match]}" for match, _, _ in matches]
        nl_char = "\n"
        return f"Did you mean one of : \n {nl_char.join(suggestions)}"
    else:
        return f"Unknown site: {site_name}"


def verify_site(site: str) -> str:
    """Check if the passed site is a valid one and returns the three
    letter site code if found. Otherwise we use fuzzy text matching to suggest
    sites with similar names.

    Args:
        site: Three letter site code or site name
    Returns:
        str: Verified three letter site code if valid site
    """
    from openghg.types import InvalidSiteError
    from openghg.util import load_json, remove_punctuation

    site_data = load_json("site_lookup.json")

    if site.upper() in site_data:
        return site.lower()
    else:
        site = remove_punctuation(site)
        name_lookup: Dict[str, str] = {value["short_name"]: code for code, value in site_data.items()}

        try:
            return name_lookup[site].lower()
        except KeyError:
            long_names = {value["long_name"]: code for code, value in site_data.items()}
            message = find_matching_site(site_name=site, possible_sites=long_names)
            raise InvalidSiteError(message)


def multiple_inlets(site: str) -> bool:
    """Check if the passed site has more than one inlet

    Args:
        site: Three letter site code
    Returns:
        bool: True if multiple inlets
    """
    from openghg.util import load_json

    site_data = load_json("acrg_site_info.json")

    site = site.upper()
    network = next(iter(site_data[site]))

    try:
        heights = set(site_data[network]["height"])
    except KeyError:
        try:
            heights = set(site_data[network]["height_name"])
        except KeyError:
            return True

    return len(heights) > 1



# from copy import copy as _copy

# from Acquire.Service import Service as _Service

# __all__ = ["HugsService"]


# class HugsService(_Service):
#     """This is a specialisation of Service for Hugs Services"""

#     def __init__(self, other=None):
#         if isinstance(other, _Service):
#             self.__dict__ = _copy(other.__dict__)

#             if self.service_type() != "hugs":
#                 from openghg.service import HugsServiceError

#                 raise HugsServiceError(
#                     "Cannot construct a HugsService from a service which is not a hugs service!"
#                 )
#         else:
#             _Service.__init__(self)

#     def _call_local_function(self, function, args):
#         """Internal function called to short-cut local 'remote'
#            function calls
#         """
#         from hugs_service.route import hugs_functions as _hugs_functions
#         from admin.handler import create_handler as _create_handler

#         handler = _create_handler(_hugs_functions)
#         return handler(function=function, args=args)



"""
Some functions for hashing data or strings for idendification of sources
"""
import hashlib
from pathlib import Path
from typing import Dict


def hash_string(to_hash: str) -> str:
    """Return the SHA-1 hash of a string

    Args:
        to_hash: String to hash
    Returns:
        str: SHA1 hash of string
    """
    return hashlib.sha1(str(to_hash).encode("utf-8")).hexdigest()


def hash_file(filepath: Path) -> str:
    """Opens the file at filepath and calculates its SHA1 hash

    Taken from https://stackoverflow.com/a/22058673

    Args:
        filepath (pathlib.Path): Path to file
    Returns:
        str: SHA1 hash
    """
    # Let's read stuff in 64kB chunks
    BUF_SIZE = 65536
    sha1 = hashlib.sha1()

    with open(filepath, "rb") as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha1.update(data)

    return sha1.hexdigest()


def hash_bytes(data: bytes) -> str:
    """Calculate the SHA1 sum of some data

    Args:
        data: Binary data
    Returns:
        str: SHA1 hash
    """
    return hashlib.sha1(data).hexdigest()


def hash_retrieved_data(to_hash: Dict[str, Dict]) -> Dict:
    """Hash data retrieved from a data platform. This calculates the SHA1 of the metadata
    and the start date, end date and the number of timestamps in the Dataset.

    Args:
        to_hash: Dictionary to hash
        We expected this to be a dictionary such as
        {species_key: {"data": xr.Dataset, "metadata": {...}}}
    Returns:
        dict: Dictionary of hash: species_key
    """
    from hashlib import sha1
    from json import dumps

    from openghg.util import timestamp_now

    current_timestamp = str(timestamp_now())
    hashes: Dict[str, Dict] = {}
    for key, data in to_hash.items():
        metadata = data["metadata"].copy()

        try:
            del metadata["file_created"]
        except KeyError:
            pass

        metadata_hash = sha1(dumps(metadata, sort_keys=True).encode("utf8")).hexdigest()

        ds = data["data"]

        start_date = str(ds.time.min())
        end_date = str(ds.time.max())
        n_timestamps = str(ds.time.size)

        basic_info = f"{start_date}_{end_date}_{n_timestamps}".encode("utf8")
        time_hash = sha1(basic_info).hexdigest()

        combo = (metadata_hash + time_hash).encode("utf8")
        combo_hash = sha1(combo).hexdigest()

        hashes[combo_hash] = {key: current_timestamp}

    return hashes



from __future__ import annotations

from typing import TYPE_CHECKING, List, Union

from openghg.store.base import BaseStore
from pandas import Timestamp

if TYPE_CHECKING:
    from openghg.dataobjects import METData


class METStore(BaseStore):
    """Controls the storage and retrieveal of meteorological data.

    Currently met data is retrieved from the ECMWF Copernicus data storage
    archive and then cached locally.
    """

    _root = "METStore"
    _uuid = "9fcabd0c-9b68-4ab4-a116-bc30a4472d67"

    def save(self) -> None:
        """Save the object to the object store

        Args:
            bucket: Bucket for data
        Returns:
            None
        """
        from openghg.objectstore import get_bucket, set_object_from_json

        bucket = get_bucket()

        obs_key = f"{METStore._root}/uuid/{METStore._uuid}"

        self._stored = True
        set_object_from_json(bucket=bucket, key=obs_key, data=self.to_data())

    @staticmethod
    def retrieve(site: str, network: str, years: Union[str, List[str]]) -> METData:
        """Retrieve data from either the local METStore or from a
        remote store if we don't have it locally

        Args:
            site: Three letter site code
            network: Network name
            years: Year(s) required
        Returns:
            METData: METData object holding data and metadata
        """
        from openghg.retrieve.met import retrieve_met
        from pandas import Timestamp

        if not isinstance(years, list):
            years = [years]
        else:
            years = sorted(years)

        store = METStore.load()

        # We'll just do full years for now, I don't think it's a huge amount of data (currently)
        start_date = Timestamp(f"{years[0]}-1-1")
        end_date = Timestamp(f"{years[-1]}-12-31")

        result = store.search(site=site, network=network, start_date=start_date, end_date=end_date)

        # Retrieve from the Copernicus store
        if result is None:
            result = retrieve_met(site=site, network=network, years=years)

            store._store(met_data=result)

        return result

    def search(
        self,
        site: str,
        network: str,
        start_date: Union[str, Timestamp],
        end_date: Union[str, Timestamp],
    ) -> Union[METData, None]:
        """Search the stored MET data

        Args:
            site: Site code
            network: Network name
            start_date: Start date
            end_date: End date

        Returns:
            METData or None: METData object if found else None
        """
        from openghg.dataobjects import METData
        from openghg.store.base import Datasource

        datasources = (Datasource.load(uuid=uuid, shallow=True) for uuid in self._datasource_uuids)

        # We should only get one datasource here currently
        for datasource in datasources:
            if datasource.search_metadata(site=site, network=network, find_all=True):
                if datasource.in_daterange(start_date=start_date, end_date=end_date):
                    data = next(iter(datasource.data().values()))
                    return METData(data=data, metadata=datasource.metadata())

        return None

    def _store(self, met_data: METData) -> None:
        """Store MET data within a Datasource

        Here we do some retrieve on the request JSON to
        make the metadata more easily searchable and of a similar
        format to Datasources used in other modules of OpenGHG.

        Args:
            met_data: Dataset
        Returns:
            None
        """
        from openghg.store.base import Datasource

        metadata = met_data.metadata

        datasource = Datasource()
        datasource.add_data(metadata=metadata, data=met_data.data, data_type="met")
        datasource.save()

        date_str = f"{metadata['start_date']}_{metadata['end_date']}"

        name = "_".join((metadata["site"], metadata["network"], date_str))
        self._datasource_uuids[datasource.uuid()] = name
        # Write this updated object back to the object store
        self.save()



from pathlib import Path
from typing import Dict, List, Optional, Union

from openghg.store import ObsSurface


def add_noaa_obspack(
    data_directory: Union[str, Path], project: Optional[str] = None, overwrite: bool = False
) -> Dict:
    """
    Function to detect and add files from the NOAA ObsPack to the object store.

    Args:
        data_directory: Top level directory for the downloaded NOAA ObsPack
        project (optional) : Can specify project or type to process only e.g. "surface"
        or "surface-flask"
        overwrite : Whether to overwrite existing entries in the object store
    Returns:
        Dict: Details of data which has been processed into the object store
    Examples:
        To add all NOAA ObsPack data (which can be processed) to the object store:
        >>> add_noaa_obspack(Path("/home/user/obspack_ch4_1_GLOBALVIEWplus_v2.0_2020-04-24"))
        {"processed": {"ch4_esp_surface-flask_2_representative.nc":{"ch4": ...}, ...}}

        To add NOAA ObsPack data for one type e.g. "surface":
        >>> add_noaa_obspack(Path("/home/user/obspack_ch4_1_GLOBALVIEWplus_v2.0_2020-04-24"), "surface")
        {"processed": {"ch4_esp_surface-flask_2_representative.nc":{"ch4": ...}, ...}}

        To add NOAA ObsPack data for one project type e.g. "surface-flask"
        >>> add_noaa_obspack(Path("/home/user/obspack_ch4_1_GLOBALVIEWplus_v2.0_2020-04-24"), "surface-flask")
        {"processed": {"ch4_esp_surface-flask_2_representative.nc":{"ch4": ...}, ...}}

    TODO: At the moment this will exclude all types which we can't process
    yet e.g. aircraft, shipboard, aircorenoaa. These should be updated once
    this functionality has been added.
    """

    # Options which we can process at the moment (ObsSurface)
    project_options = {"surface": ["flask", "insitu", "pfp"], "tower": ["insitu"]}

    project_names = _create_project_names(project_options)

    # Options we can't process at the moment but may be encountered (ObsMobile, ...).
    # TODO: "tower-insitu" should be able to run through ObsSurface but
    # these contain multiple heights per file - not sure we are handling this yet
    project_options_not_implemented_yet = {
        "aircraft": ["pfp", "insitu"],
        "shipboard": ["flask"],
        "aircorenoaa": [""],
    }

    project_names_not_implemented = _create_project_names(project_options_not_implemented_yet)

    # If a specific project has been specified, extract file matching strings
    if project:
        if project in project_names:
            projects_to_read = [project]
        elif project in project_options:
            projects_to_read = [name for name in project_names if project in name]
        elif project in project_options_not_implemented_yet or project in project_names_not_implemented:
            raise ValueError(f"Functionality to process {project} data has not been implemented yet.")
        else:
            raise ValueError(f"Did not recognise input {project} for project")
    else:
        projects_to_read = project_names
        print(f"Reading subset of data from {','.join(projects_to_read)}")

    if isinstance(data_directory, str):
        data_directory = Path(data_directory)

    # ObsPack may contain nc or txt files
    # Try and extract one and then the other if possible
    files = _find_noaa_files(data_directory, ".nc")
    if not files:
        files = _find_noaa_files(data_directory, ".txt")

    # Find relevant details for each file and call parse_noaa() function
    processed_summary: Dict[str, Dict] = {}
    for filepath in files:
        param = _param_from_filename(filepath)
        site = param["site"]
        project = param["project"]
        measurement_type = param["measurement_type"]

        if project in projects_to_read:
            processed = ObsSurface.read_file(
                filepath,
                site=site,
                measurement_type=measurement_type,
                network="NOAA",
                data_type="NOAA",
                overwrite=overwrite,
            )
        elif project in project_names_not_implemented:
            print(f"Not processing {filepath.name} - no standardisation for {project} data implemented yet.")
            processed = {}
        else:
            processed = {}

        # Expect "processed" dictionary and/or "error" dictionary within `processed`
        for key, value in processed.items():
            if key not in processed_summary:
                processed_summary[key] = {}
            for key_in, value_in in value.items():
                processed_summary[key][key_in] = value_in

    return processed_summary


def _param_from_filename(filename: Union[str, Path]) -> Dict:
    """
    Extract parameter from the NOAA filename based on the agreed naming convention.
    See: https://gml.noaa.gov/ccgg/obspack/documentation.html

    <trace gas identifier>_<site code>_<project>_<lab number>_<selection tag>.<filetype extension>

    Args:
        filename: NOAA ObsPack filename with expected naming convention
    Returns:
        Dict : Extracted parameters from filename
    Examples:
        >>> _param_from_filename("ch4_esp_surface-flask_2_representative.nc")
        {"species": "ch4", "site" : "esp", "project": "surface-flask", "measurement_type": "flask"}
    """

    if isinstance(filename, str):
        extracted_param = filename.split("_")
    else:
        extracted_param = filename.name.split("_")
    param = {}
    param["species"] = extracted_param[0]
    param["site"] = extracted_param[1]
    param["project"] = extracted_param[2]
    param["measurement_type"] = param["project"].split("-")[-1]

    return param


def _create_project_names(input_dict: Dict) -> List:
    """
    Creates full project names as would be included in the NOAA filepath

    Expects input dictionary for each the type e.g. "surface" and the
    associated measurement types e.g. ["flask", "insitu", "pfp"]

    Args:
        input_dict: In format described above
    Returns:
        List: collated names for each project
    Examples:
        The input dictionary should be defined as follows, where each value is a list.
        >>> input_dict = {"surface": ["flask", "insitu", "pfp"]}
        >>> _create_project_names(input_dict)
           ["surface-flask", "surface-insitu", "surface-php"]
    """
    projects = []
    for key, values in input_dict.items():
        if not isinstance(values, list):
            values = [values]
        full_names = ["-".join([key, value]) for value in values]
        projects.extend(full_names)
    return projects


def _find_noaa_files(data_directory: Union[str, Path], ext: str) -> List:
    """
    Find obs files in NOAA ObsPack.

    Expected directory structure is:
     - <ObsPack>/data/<filetype>/
       - e.g. obspack_ch4_1_GLOBALVIEWplus_v2.0_2020-04-24/data/nc/

    Args:
        data_directory: Top level ObsPack data directory to search
        ext: Extension for files. Should be either ".txt" or ".nc"
    Returns:
        list: Filenames with appropriate extension
    Examples:
        Can include top level directory (str or Path):
        >>> _find_noaa_files(Path("/home/user/obspack_ch4_1_GLOBALVIEWplus_v2.0_2020-04-24"), ".txt")
        ["co_pocn25_surface-flask_1_ccgg_event.txt", ...]

        Or optionally can include more direct subdirectory (for nc files):
        >>> _find_noaa_files("/home/user/obspack_ch4_1_GLOBALVIEWplus_v2.0_2020-04-24/data/nc", ".nc")
        ["ch4_esp_surface-flask_2_representative.nc", ...]
    """

    # ObsPack may contain nc or txt files:
    # - For nc files found, these should all the data files
    # - For txt files found, we need to make sure files are found in the correct
    # sub-directory as otherwise this may find README and summary files
    if ext == ".nc":
        subdirectories = ["data/nc", "nc", ""]
    elif ext == ".txt":
        subdirectories = ["data/txt", "txt"]
    else:
        raise ValueError("Did not recognise input for extension: {ext}. Should be one of '.txt' or '.nc'")

    data_directory = Path(data_directory).expanduser().resolve()

    # Allow user to specify various levels within ObsPack to e.g. just
    # extract nc files.
    for subdir in subdirectories:
        path_to_search = data_directory / subdir
        print(f"Searching for {ext} files within {path_to_search}")
        files = list(path_to_search.glob(f"*{ext}"))
        suffix_values = [file.suffix for file in files]
        if ext in suffix_values:
            break
    else:
        files = []

    return files



from ._enum import DataTypes, EmissionsDatabases, EmissionsTypes, ObsTypes, SurfaceTypes
from ._errors import (
    DatasourceLookupError,
    FunctionError,
    InvalidSiteError,
    ObjectStoreError,
    UnknownDataError,
)
from ._types import multiPathType, pathType, resultsType



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, Optional, Tuple, Union

import numpy as np
from numpy import ndarray
from openghg.store import DataSchema
from openghg.store.base import BaseStore
from xarray import DataArray, Dataset

__all__ = ["Emissions"]


ArrayType = Optional[Union[ndarray, DataArray]]


class Emissions(BaseStore):
    """This class is used to process emissions / flux data"""

    _root = "Emissions"
    _uuid = "c5c88168-0498-40ac-9ad3-949e91a30872"
    _metakey = f"{_root}/uuid/{_uuid}/metastore"

    @staticmethod
    def read_data(binary_data: bytes, metadata: Dict, file_metadata: Dict) -> Dict:
        """Ready a footprint from binary data

        Args:
            binary_data: Footprint data
            metadata: Dictionary of metadata
            file_metadat: File metadata
        Returns:
            dict: UUIDs of Datasources data has been assigned to
        """
        with TemporaryDirectory() as tmpdir:
            tmpdir_path = Path(tmpdir)

            try:
                filename = file_metadata["filename"]
            except KeyError:
                raise KeyError("We require a filename key for metadata read.")

            filepath = tmpdir_path.joinpath(filename)
            filepath.write_bytes(binary_data)

            return Emissions.read_file(filepath=filepath, **metadata)

    @staticmethod
    def read_file(
        filepath: Union[str, Path],
        species: str,
        source: str,
        domain: str,
        data_type: str = "openghg",
        date: Optional[str] = None,
        high_time_resolution: Optional[bool] = False,
        period: Optional[Union[str, tuple]] = None,
        continuous: bool = True,
        overwrite: bool = False,
    ) -> Dict:
        """Read emissions file

        Args:
            filepath: Path of emissions file
            species: Species name
            domain: Emissions domain
            source: Emissions source
            date : Date associated with emissions as a string
            data_type : Type of data being input e.g. openghg (internal format)
            high_time_resolution: If this is a high resolution file
            period: Period of measurements. Only needed if this can not be inferred from the time coords
                    If specified, should be one of:
                     - "yearly", "monthly"
                     - suitable pandas Offset Alias
                     - tuple of (value, unit) as would be passed to pandas.Timedelta function
            continuous: Whether time stamps have to be continuous.
            overwrite: Should this data overwrite currently stored data.
        Returns:
            dict: Dictionary of datasource UUIDs data assigned to
        """
        from openghg.store import assign_data, datasource_lookup, load_metastore
        from openghg.types import EmissionsTypes
        from openghg.util import clean_string, hash_file, load_emissions_parser

        species = clean_string(species)
        source = clean_string(source)
        domain = clean_string(domain)
        date = clean_string(date)

        filepath = Path(filepath)

        try:
            data_type = EmissionsTypes[data_type.upper()].value
        except KeyError:
            raise ValueError(f"Unknown data type {data_type} selected.")

        # Load the data retrieve object
        parser_fn = load_emissions_parser(data_type=data_type)

        em_store = Emissions.load()

        # Load in the metadata store
        metastore = load_metastore(key=em_store._metakey)

        file_hash = hash_file(filepath=filepath)
        if file_hash in em_store._file_hashes and not overwrite:
            print(
                f"This file has been uploaded previously with the filename : {em_store._file_hashes[file_hash]} - skipping."
            )

        # Define parameters to pass to the parser function
        # TODO: Update this to match against inputs for parser function.
        param = {
            "filepath": filepath,
            "species": species,
            "domain": domain,
            "source": source,
            "date": date,
            "high_time_resolution": high_time_resolution,
            "period": period,
            "continuous": continuous,
        }

        emissions_data = parser_fn(**param)

        # Checking against expected format for Emissions
        for split_data in emissions_data.values():
            em_data = split_data["data"]
            Emissions.validate_data(em_data)

        required = ("species", "source", "domain", "date")
        lookup_results = datasource_lookup(metastore=metastore, data=emissions_data, required_keys=required)

        data_type = "emissions"
        datasource_uuids = assign_data(
            data_dict=emissions_data,
            lookup_results=lookup_results,
            overwrite=overwrite,
            data_type=data_type,
        )

        em_store.add_datasources(uuids=datasource_uuids, data=emissions_data, metastore=metastore)

        # Record the file hash in case we see this file again
        em_store._file_hashes[file_hash] = filepath.name

        em_store.save()
        metastore.close()

        return datasource_uuids

    @staticmethod
    def transform_data(
        datapath: Union[str, Path],
        database: str,
        overwrite: bool = False,
        **kwargs: Dict,
    ) -> Dict:
        """
        Read and transform an emissions database. This will find the appropriate
        parser function to use for the database specified. The necessary inputs
        are determined by which database ie being used.

        The underlying parser functions will be of the form:
            - openghg.transform.emissions.parse_{database.lower()}
                - e.g. openghg.transform.emissions.parse_edgar()

        Args:
            datapath: Path to local copy of database archive (for now)
            database: Name of database
            overwrite: Should this data overwrite currently stored data
                which matches.
            **kwargs: Inputs for underlying parser function for the database.
                Necessary inputs will depend on the database being parsed.

        TODO: Could allow Callable[..., Dataset] type for a pre-defined function be passed
        """
        import inspect

        from openghg.store import assign_data, datasource_lookup, load_metastore
        from openghg.types import EmissionsDatabases
        from openghg.util import load_emissions_database_parser

        datapath = Path(datapath)

        try:
            data_type = EmissionsDatabases[database.upper()].value
        except KeyError:
            raise ValueError(f"Unable to transform '{database}' selected.")

        # Load the data retrieve object
        parser_fn = load_emissions_database_parser(database=database)

        em_store = Emissions.load()

        # Load in the metadata store
        metastore = load_metastore(key=em_store._metakey)

        # Find all parameters that can be accepted by parse function
        all_param = list(inspect.signature(parser_fn).parameters.keys())

        # Define parameters to pass to the parser function from kwargs
        param: Dict[Any, Any] = {key: value for key, value in kwargs.items() if key in all_param}
        param["datapath"] = datapath  # Add datapath explicitly (for now)

        emissions_data = parser_fn(**param)

        # Checking against expected format for Emissions
        for split_data in emissions_data.values():
            em_data = split_data["data"]
            Emissions.validate_data(em_data)

        required = ("species", "source", "domain", "date")
        lookup_results = datasource_lookup(metastore=metastore, data=emissions_data, required_keys=required)

        data_type = "emissions"
        overwrite = False
        datasource_uuids = assign_data(
            data_dict=emissions_data,
            lookup_results=lookup_results,
            overwrite=overwrite,
            data_type=data_type,
        )

        em_store.add_datasources(uuids=datasource_uuids, data=emissions_data, metastore=metastore)

        em_store.save()
        metastore.close()

        return datasource_uuids

    @staticmethod
    def schema() -> DataSchema:
        """
        Define schema for emissions Dataset.

        Includes flux/emissions for each time and position:
            - "flux"
                - expected dimensions: ("time", "lat", "lon")

        Expected data types for all variables and coordinates also included.

        Returns:
            DataSchema : Contains schema for Emissions.
        """
        data_vars: Dict[str, Tuple[str, ...]] = {"flux": ("time", "lat", "lon")}
        dtypes = {"lat": np.floating, "lon": np.floating, "time": np.datetime64, "flux": np.floating}

        data_format = DataSchema(data_vars=data_vars, dtypes=dtypes)

        return data_format

    @staticmethod
    def validate_data(data: Dataset) -> None:
        """
        Validate input data against Emissions schema - definition from
        Emissions.schema() method.

        Args:
            data : xarray Dataset in expected format

        Returns:
            None

            Raises a ValueError with details if the input data does not adhere
            to the Emissions schema.
        """
        data_schema = Emissions.schema()
        data_schema.validate_data(data)

    def lookup_uuid(self, species: str, source: str, domain: str, date: str) -> Union[str, bool]:
        """Perform a lookup for the UUID of a Datasource

        Args:
            species: Site code
            domain: Domain
            model: Model name
            height: Height
        Returns:
            str or dict: UUID or False if no entry
        """
        uuid = self._datasource_table[species][source][domain][date]

        return uuid if uuid else False

    def set_uuid(self, species: str, source: str, domain: str, date: str, uuid: str) -> None:
        """Record a UUID of a Datasource in the datasource table

        Args:
            site: Site code
            domain: Domain
            model: Model name
            height: Height
            uuid: UUID of Datasource
        Returns:
            None
        """
        self._datasource_table[species][source][domain][date] = uuid



import json
from io import BytesIO
from typing import Dict, List, Optional, Union

from openghg.dataobjects import BoundaryConditionsData, FluxData, FootprintData, ObsData
from openghg.util import decompress, decompress_str, hash_bytes, running_on_hub
from pandas import Timestamp
from xarray import Dataset, load_dataset


def get_obs_surface(
    site: str,
    species: str,
    inlet: str = None,
    start_date: Union[str, Timestamp] = None,
    end_date: Union[str, Timestamp] = None,
    average: str = None,
    network: str = None,
    instrument: str = None,
    calibration_scale: str = None,
    keep_missing: bool = False,
    skip_ranking: bool = False,
) -> Optional[ObsData]:
    """This is the equivalent of the get_obs function from the ACRG repository.

    Usage and return values are the same whilst implementation may differ.

    Args:
        site: Site of interest e.g. MHD for the Mace Head site.
        species: Species identifier e.g. ch4 for methane.
        start_date: Output start date in a format that Pandas can interpret
        end_date: Output end date in a format that Pandas can interpret
        inlet: Inlet label
        average: Averaging period for each dataset. Each value should be a string of
        the form e.g. "2H", "30min" (should match pandas offset aliases format).
        keep_missing: Keep missing data points or drop them.
        network: Network for the site/instrument (must match number of sites).
        instrument: Specific instrument for the sipte (must match number of sites).
        calibration_scale: Convert to this calibration scale
    Returns:
        ObsData or None: ObsData object if data found, else None
    """
    from openghg.cloud import call_function

    if running_on_hub():
        to_post: Dict[str, Union[str, Dict]] = {}

        to_post["function"] = "get_obs_surface"

        search_terms = {
            "site": site,
            "species": species,
            "keep_missing": keep_missing,
            "skip_ranking": skip_ranking,
        }

        if inlet is not None:
            search_terms["inlet"] = inlet
        if start_date is not None:
            search_terms["start_date"] = start_date
        if end_date is not None:
            search_terms["end_date"] = end_date
        if average is not None:
            search_terms["average"] = average
        if network is not None:
            search_terms["network"] = network
        if instrument is not None:
            search_terms["instrument"] = instrument
        if calibration_scale is not None:
            search_terms["calibration_scale"] = calibration_scale

        to_post["search_terms"] = search_terms

        result = call_function(data=to_post)

        content = result["content"]
        found = content["found"]

        if found:
            binary_data = decompress(data=content["data"])

            file_metadata = content["file_metadata"]
            sha1_hash_data = file_metadata["data"]["sha1_hash"]

            if sha1_hash_data != hash_bytes(data=binary_data):
                raise ValueError("Hash mismatch between local SHA1 and remote SHA1.")

            buf = BytesIO(binary_data)
            json_str = decompress_str(data=content["metadata"])
            metadata = json.loads(json_str)
            dataset = load_dataset(buf)

            return ObsData(data=dataset, metadata=metadata)
        else:
            return None
    else:
        return get_obs_surface_local(
            site=site,
            species=species,
            start_date=start_date,
            end_date=end_date,
            inlet=inlet,
            average=average,
            network=network,
            instrument=instrument,
            calibration_scale=calibration_scale,
            keep_missing=keep_missing,
            skip_ranking=skip_ranking,
        )


def get_obs_surface_local(
    site: str,
    species: str,
    inlet: Optional[str] = None,
    start_date: Optional[Union[str, Timestamp]] = None,
    end_date: Optional[Union[str, Timestamp]] = None,
    average: Optional[str] = None,
    network: Optional[str] = None,
    instrument: Optional[str] = None,
    calibration_scale: Optional[str] = None,
    keep_missing: Optional[bool] = False,
    skip_ranking: Optional[bool] = False,
) -> Optional[ObsData]:
    """This is the equivalent of the get_obs function from the ACRG repository.

    Usage and return values are the same whilst implementation may differ.

    This function should not be used on the OpenGHG Hub.

    Args:
        site: Site of interest e.g. MHD for the Mace Head site.
        species: Species identifier e.g. ch4 for methane.
        start_date: Output start date in a format that Pandas can interpret
        end_date: Output end date in a format that Pandas can interpret
        inlet: Inlet label
        average: Averaging period for each dataset. Each value should be a string of
        the form e.g. "2H", "30min" (should match pandas offset aliases format).
        keep_missing: Keep missing data points or drop them.
        network: Network for the site/instrument (must match number of sites).
        instrument: Specific instrument for the sipte (must match number of sites).
        calibration_scale: Convert to this calibration scale
    Returns:
        ObsData or None: ObsData object if data found, else None
    """
    import numpy as np
    from openghg.retrieve import search_surface
    from openghg.store import recombine_datasets
    from openghg.util import clean_string, load_json, synonyms, timestamp_tzaware
    from pandas import Timedelta, Timestamp
    from xarray import concat as xr_concat

    if running_on_hub():
        raise ValueError(
            "This function cannot be used on the OpenGHG Hub. Please use openghg.retrieve.get_obs_surface instead."
        )

    site_info = load_json(filename="acrg_site_info.json")
    site = site.upper()

    if site not in site_info:
        raise ValueError(f"No site called {site}, please enter a valid site name.")

    # Find the correct synonym for the passed species
    species = clean_string(synonyms(species))

    # Get the observation data
    obs_results = search_surface(
        site=site,
        species=species,
        inlet=inlet,
        start_date=start_date,
        end_date=end_date,
        instrument=instrument,
        find_all=True,
        skip_ranking=skip_ranking,
    )

    if not obs_results:
        raise ValueError(f"Unable to find results for {species} at {site}")

    retrieved_data: Union[ObsData, List[ObsData], None] = obs_results.retrieve(
        site=site, species=species, inlet=inlet
    )

    if retrieved_data is None:
        return None
    elif isinstance(retrieved_data, list):
        print("No data returned.")
        print(f"Multiple entries found for current input parameters - site: '{site}', species: '{species}'")
        print("Please supply additional parameters or set ranking.")
        metadata_difference(retrieved_data, params=["inlet", "network", "instrument"])
        return None

    data = retrieved_data.data

    if data.attrs["inlet"] == "multiple":
        data.attrs["inlet_height_magl"] = "multiple"
        retrieved_data.metadata["inlet"] = "multiple"

    if start_date is not None and end_date is not None:
        start_date_tzaware = timestamp_tzaware(start_date)
        end_date_tzaware = timestamp_tzaware(end_date)
        end_date_tzaware_exclusive = end_date_tzaware - Timedelta(
            1, unit="nanosecond"
        )  # Deduct 1 ns to make the end day (date) exclusive.

        # Slice the data to only cover the dates we're interested in
        data = data.sel(time=slice(start_date_tzaware, end_date_tzaware_exclusive))

    try:
        start_date_data = timestamp_tzaware(data.time[0].values)
        end_date_data = timestamp_tzaware(data.time[-1].values)
    except AttributeError:
        raise AttributeError("This dataset does not have a time attribute, unable to read date range")
    except IndexError:
        return None

    if average is not None:
        # GJ - 2021-03-09
        # TODO - check by RT

        # # Average the Dataset over a given period
        # if keep_missing is True:
        #     # Create a dataset with one element and NaNs to prepend or append
        #     ds_single_element = data[{"time": 0}]

        #     for v in ds_single_element.variables:
        #         if v != "time":
        #             ds_single_element[v].values = np.nan

        #     ds_concat = []

        #     # Pad with an empty entry at the start date
        #     if timestamp_tzaware(data.time.min()) > start_date:
        #         ds_single_element_start = ds_single_element.copy()
        #         ds_single_element_start.time.values = Timestamp(start_date)
        #         ds_concat.append(ds_single_element_start)

        #     ds_concat.append(data)

        #     # Pad with an empty entry at the end date
        #     if data.time.max() < Timestamp(end_date):
        #         ds_single_element_end = ds_single_element.copy()
        #         ds_single_element_end.time.values = Timestamp(end_date) - Timedelta("1ns")
        #         ds_concat.append(ds_single_element_end)

        #     data = xr_concat(ds_concat, dim="time")

        #     # Now sort to get everything in the right order
        #     data = data.sortby("time")

        # First do a mean resample on all variables
        ds_resampled = data.resample(time=average).mean(skipna=False, keep_attrs=True)
        # keep_attrs doesn't seem to work for some reason, so manually copy
        ds_resampled.attrs = data.attrs.copy()

        average_in_seconds = Timedelta(average).total_seconds()
        ds_resampled.attrs["averaged_period"] = average_in_seconds
        ds_resampled.attrs["averaged_period_str"] = average

        # For some variables, need a different type of resampling
        data_variables: List[str] = [str(v) for v in data.variables]

        for var in data_variables:
            if "repeatability" in var:
                ds_resampled[var] = (
                    np.sqrt((data[var] ** 2).resample(time=average).sum())
                    / data[var].resample(time=average).count()
                )

            # Copy over some attributes
            if "long_name" in data[var].attrs:
                ds_resampled[var].attrs["long_name"] = data[var].attrs["long_name"]

            if "units" in data[var].attrs:
                ds_resampled[var].attrs["units"] = data[var].attrs["units"]

        # Create a new variability variable, containing the standard deviation within the resampling period
        ds_resampled[f"{species}_variability"] = (
            data[species].resample(time=average).std(skipna=False, keep_attrs=True)
        )

        # If there are any periods where only one measurement was resampled, just use the median variability
        ds_resampled[f"{species}_variability"][ds_resampled[f"{species}_variability"] == 0.0] = ds_resampled[
            f"{species}_variability"
        ].median()

        # Create attributes for variability variable
        ds_resampled[f"{species}_variability"].attrs["long_name"] = f"{data.attrs['long_name']}_variability"

        ds_resampled[f"{species}_variability"].attrs["units"] = data[species].attrs["units"]

        # Resampling may introduce NaNs, so remove, if not keep_missing
        if keep_missing is False:
            ds_resampled = ds_resampled.dropna(dim="time")

        data = ds_resampled

    # Rename variables
    rename: Dict[str, str] = {}

    data_variables = [str(v) for v in data.variables]
    for var in data_variables:
        if var.lower() == species.lower():
            rename[var] = "mf"
        if "repeatability" in var:
            rename[var] = "mf_repeatability"
        if "variability" in var:
            rename[var] = "mf_variability"
        if "number_of_observations" in var:
            rename[var] = "mf_number_of_observations"
        if "status_flag" in var:
            rename[var] = "status_flag"
        if "integration_flag" in var:
            rename[var] = "integration_flag"

    data = data.rename_vars(rename)  # type: ignore

    data.attrs["species"] = species

    if "calibration_scale" in data.attrs:
        data.attrs["scale"] = data.attrs.pop("calibration_scale")

    if calibration_scale is not None:
        data = _scale_convert(data, species, calibration_scale)

    metadata = retrieved_data.metadata
    metadata.update(data.attrs)

    obs_data = ObsData(data=data, metadata=metadata)

    # It doesn't make sense to do this now as we've only got a single Dataset
    # # Now check if the units match for each of the observation Datasets
    # units = set((f.data.mf.attrs["units"] for f in obs_files))
    # scales = set((f.data.attrs["scale"] for f in obs_files))

    # if len(units) > 1:
    #     raise ValueError(
    #         f"Units do not match for these observation Datasets {[(f.mf.attrs['station_long_name'],f.attrs['units']) for f in obs_files]}"
    #     )

    # if len(scales) > 1:
    #     print(
    #         f"Scales do not match for these observation Datasets {[(f.mf.attrs['station_long_name'],f.attrs['units']) for f in obs_files]}"
    #     )
    #     print("Suggestion: set calibration_scale to convert scales")

    return obs_data


def get_flux(
    species: str,
    source: str,
    domain: str,
    start_date: Optional[Timestamp] = None,
    end_date: Optional[Timestamp] = None,
    time_resolution: Optional[str] = None,
) -> FluxData:
    """
    The flux function reads in all flux files for the domain and species as an xarray Dataset.
    Note that at present ALL flux data is read in per species per domain or by emissions name.
    To be consistent with the footprints, fluxes should be in mol/m2/s.

    Args:
        species: Species name
        source: Source name
        domain: Domain e.g. EUROPE
        start_date: Start date
        end_date: End date
        time_resolution: One of ["standard", "high"]
    Returns:
        FluxData: FluxData object
    """
    from openghg.retrieve import search
    from openghg.store import recombine_datasets
    from openghg.util import clean_string, synonyms, timestamp_epoch, timestamp_now

    # Find the correct synonym for the passed species
    species = clean_string(synonyms(species))

    if start_date is None:
        start_date = timestamp_epoch()
    if end_date is None:
        end_date = timestamp_now()

    results: Dict = search(
        species=species,
        source=source,
        domain=domain,
        time_resolution=time_resolution,
        start_date=start_date,
        end_date=end_date,
        data_type="emissions",
    )  # type: ignore

    if not results:
        raise ValueError(f"Unable to find flux data for {species} from {source}")

    try:
        em_key = list(results.keys())[0]
    except IndexError:
        raise ValueError(f"Unable to find any footprints data for {domain} for {species}.")

    data_keys = results[em_key]["keys"]
    metadata = results[em_key]["metadata"]

    em_ds = recombine_datasets(keys=data_keys, sort=False)

    # Check for level coordinate. If one level, assume surface and drop
    if "lev" in em_ds.coords:
        if len(em_ds.lev) > 1:
            raise ValueError("Error: More than one flux level")

        em_ds = em_ds.drop_vars(names="lev")

    if species is None:
        species = metadata.get("species", "NA")

    return FluxData(data=em_ds, metadata=metadata)


def get_bc(
    species: str,
    domain: str,
    bc_input: Optional[str] = None,
    start_date: Optional[Timestamp] = None,
    end_date: Optional[Timestamp] = None,
) -> BoundaryConditionsData:
    """
    Get boundary conditions for a given species, domain and bc_input name.

    Args:
        species: Species name
        bc_input: Input used to create boundary conditions. For example:
            - a model name such as "MOZART" or "CAMS"
            - a description such as "UniformAGAGE" (uniform values based on AGAGE average)
        domain: Region for boundary conditions e.g. EUROPE
        start_date: Start date
        end_date: End date
    Returns:
        BoundaryConditionsData: BoundaryConditionsData object
    """
    from openghg.retrieve import search
    from openghg.store import recombine_datasets
    from openghg.util import clean_string, synonyms, timestamp_epoch, timestamp_now

    # Find the correct synonym for the passed species
    species = clean_string(synonyms(species))

    if start_date is None:
        start_date = timestamp_epoch()
    if end_date is None:
        end_date = timestamp_now()

    results: Dict = search(
        species=species,
        bc_input=bc_input,
        domain=domain,
        start_date=start_date,
        end_date=end_date,
        data_type="boundary_conditions",
    )  # type: ignore

    if not results:
        raise ValueError(f"Unable to find boundary conditions data for {species} for {bc_input}")

    try:
        bc_key = list(results.keys())[0]
    except IndexError:
        raise ValueError(f"Unable to find any boundary conditions data for {domain} for {species}.")

    data_keys = results[bc_key]["keys"]
    metadata = results[bc_key]["metadata"]

    bc_ds = recombine_datasets(keys=data_keys, sort=False)

    if species is None:
        species = metadata.get("species", "NA")

    return BoundaryConditionsData(
        data=bc_ds,
        metadata=metadata,
    )


def get_footprint(
    site: str,
    domain: str,
    height: str,
    model: str = None,
    start_date: Timestamp = None,
    end_date: Timestamp = None,
    species: str = None,
) -> FootprintData:
    """
    Get footprints from one site.

    Args:
        site: The name of the site given in the footprints. This often matches
              to the site name but  if the same site footprints are run with a
              different met and they are named slightly differently from the obs
              file. E.g. site="DJI", site_modifier = "DJI-SAM" -
              station called DJI, footprints site called DJI-SAM
        domain : Domain name for the footprints
        height: Height of inlet in metres
        start_date: Output start date in a format that Pandas can interpret
        end_date: Output end date in a format that Pandas can interpret
        species: Species identifier e.g. "co2" for carbon dioxide. Only needed
                 if species needs a modified footprints from the typical 30-day
                 footprints appropriate for a long-lived species (like methane)
                 e.g. for high time resolution (co2) or is a short-lived species.
    Returns:
        FootprintData: FootprintData dataclass
    """
    from openghg.dataobjects import FootprintData
    from openghg.retrieve import search
    from openghg.store import recombine_datasets
    from openghg.util import clean_string, synonyms

    # Find the correct synonym for the passed species
    if species is not None:
        species = clean_string(synonyms(species))

    results: Dict = search(
        site=site,
        domain=domain,
        height=height,
        start_date=start_date,
        end_date=end_date,
        species=species,
        data_type="footprints",
    )  # type: ignore
    # Get the footprints data
    # if species is not None:
    # else:
    #     results = search(
    #         site=site,
    #         domain=domain,
    #         domain=domain,
    #         height=height,
    #         start_date=start_date,
    #         end_date=end_date,
    #         data_type="footprints",
    #     )  # type: ignore

    try:
        fp_site_key = list(results.keys())[0]
    except IndexError:
        if species is not None:
            raise ValueError(
                f"Unable to find any footprints data for {site} at a height of {height} for species {species}."
            )
        else:
            raise ValueError(f"Unable to find any footprints data for {site} at a height of {height}.")

    keys = results[fp_site_key]["keys"]
    metadata = results[fp_site_key]["metadata"]
    # fp_ds = recombine_datasets(keys=keys, sort=False) # Why did this have sort=False before?
    fp_ds = recombine_datasets(keys=keys, sort=True)

    # TODO: Could incorporate this somewhere? Setting species to INERT?
    # if species is None:
    #     species = metadata.get("species", "INERT")

    return FootprintData(data=fp_ds, metadata=metadata)


def _scale_convert(data: Dataset, species: str, to_scale: str) -> Dataset:
    """Convert to a new calibration scale

    Args:
        data: Must contain an mf variable (mole fraction), and scale must be in global attributes
        species: species name
        to_scale: Calibration scale to convert to
    Returns:
        xarray.Dataset: Dataset with mole fraction data scaled
    """
    from numexpr import evaluate
    from openghg.util import get_datapath
    from pandas import read_csv

    # If scale is already correct, return
    ds_scale = data.attrs["scale"]
    if ds_scale == to_scale:
        return data

    scale_convert_filepath = get_datapath("acrg_obs_scale_convert.csv")

    scale_converter = read_csv(scale_convert_filepath)
    scale_converter_scales = scale_converter[scale_converter.isin([species.upper(), ds_scale, to_scale])][
        ["species", "scale1", "scale2"]
    ].dropna(axis=0, how="any")

    if len(scale_converter_scales) == 0:
        raise ValueError(
            f"Scales {ds_scale} and {to_scale} are not both in any one row in acrg_obs_scale_convert.csv for species {species}"
        )
    elif len(scale_converter_scales) > 1:
        raise ValueError(f"Duplicate rows in acrg_obs_scale_convert.csv?")
    else:
        row = scale_converter_scales.index[0]

    converter = scale_converter.loc[row]

    if to_scale == converter["scale1"]:
        direction = "2to1"
    else:
        direction = "1to2"

    # flake8: noqa: F841
    # scale_convert file has variable X in equations, so let's create it
    X = 1.0
    scale_factor = evaluate(converter[direction])
    data["mf"].values *= scale_factor

    data.attrs["scale"] = to_scale

    return data


multDataTypes = Union[List[ObsData], List[FootprintData], List[FluxData]]


def metadata_difference(
    data: multDataTypes, params: Optional[list] = None, print_output: bool = True
) -> list:
    """
    Check differences between metadata for returned data objects.

    Args:
        data : Multiple data objects e.g. multiple ObsData as a list
        params : Specific metadata parameters to check. If None all parameters will be checked
        print_output : Summarise and print output to screen.

    Returns:
        list : Keys from the metadata with differences
    """
    metadata = [d.metadata for d in data]
    if params is not None:
        metadata = [{param: m[param] for param in params} for m in metadata]

    metadata0 = metadata[0]
    difference = []
    for metadata_compare in metadata[1:]:
        metadata_diff = set(metadata0.items()) - set(metadata_compare.items())
        difference.extend(list(metadata_diff))
    param_difference = list(set([d[0] for d in difference]))

    ignore_params = ["data_owner", "data_owner_email"]
    for iparam in ignore_params:
        try:
            param_difference.remove(iparam)
        except ValueError:
            continue

    if print_output:
        print("Datasets contain:")
        for param in param_difference:
            print(f" {param}: ", end="")
            for m in metadata:
                print(f" '{m[param]}', ", end="")
            print()  # print new line

    return param_difference
